        subroutine update_Ed(ntimes)
c*****************************************************************************
c       Update updates the Hubbard-Statonovich fields by sweeping sequentially 
c       through them and testing if changes are required.  The Ising field is
c       updated by the metropolis or heat bath algorithm.  If changes are made 
c       it updates the greens functions appropriately.  This block should 
c       probably be compiled with an inline option.
c       
c       This version is based on Eduardo DAzevedos proposed changes 
c       to delay the dger updates and perform a slender rectangular dgemm
c       update.  This method takes the usual Hirsch-Fye algorithm which
c       sweeps sequentially through blocks of HHS fields  and makes local
c       changes in the Green functions using 
c
c
c       G' = G + [G - 1][exp(V'-V) - 1]G'                       (1)
c
c       or 
c           -1
c       G G'   = 1 + [1-G][exp(V'-V) - 1]                       (2)
c
c       The matrix [exp(V'-V) - 1] is diagonal with only 1 finite 
c       element, and V=alpha sigma s.  If we make a local change in the
c       HHS field, s_n-->s_n', then we may iterate these equations twice
c       to obtain an updating equation for G
c
c                                                    -1 
c       G'_ij = G_ij + [G - 1]_in [exp(V'-V) - 1]_n M   G_nj        
c
c
c       M = [exp(V'-V) - 1]_n / [1 + [1-G]_nn [exp(V'-V) - 1]_n ]        (3)
c
c       This first equation is a vector outer product followed by a
c       matrix update, and may be completed with a call to dger.
c       
c       However, dger is very innefficient, since it has only two 
c       floating point operations per memory access (and both FPO
c       can be done on the same clock cycle).  Thus it is limited 
c       by the speed of the memory bus, how fast G can be moved in
c       and out of memory.
c
c       Ed proposes that we delay the Green function update, until a 
c       sufficiently large number of changes (maxstep=32) have been
c       accumulated.  Only afterwards is the whole Green function matrix 
c       updated.  If we label these 32 accepted changes sequentially, 
c       then Eq. 3 becomes 
c
c        32      1        31                          31 -1 31
c       G_ ij = G_ ij + [G - 1]_in [exp(V'-V) - 1]_n M     G_ nj     
c
c                 30                          30 -1 30
c             + [G - 1]_in [exp(V'-V) - 1]_n M     G_ nj     
c             .
c             .
c             .
c                 1                           1 -1  1
c             + [G - 1]_in [exp(V'-V) - 1]_n M     G_ nj             (4)
c
c       This can be collaped into a matrix matrix product (nt by 32
c       times 32 by nt matrices); however, this is difficult since
c       each term depends on the prior one.  Ed noted that it only takes
c       of O(32*nt) operations to acount for these dependencies, so a 
c       great increase in speed may be obtained by rewriting Eq.4 as
c       a matrix-matrix product.
c
c        32   1
c       G  = G  + A * B 
c
c       The reason we only update 32 spins at a time, is that for
c       each spin we update, we must calculate a column and row of 
c       G to form A and B, respectively.  If we increase the number
c       of accepted changes, then we waste all of this work.  I.e.
c       the amount of cpu time required to track the dependences 
c       noted above will eventually approach nt^2, the same amount 
c       of work required to update G.  So, as maxstep-->32, this process
c       is likely to become (slightly) less efficient.
c

c
c*****************************************************************************
        use Global
        implicit none
        include 'mpif.h'
#define SIMPLE_SPRNG
#define FLOAT_GEN
#define USE_MPI
#include "sprng_f.h" 
c************************************************************
        integer, parameter :: maxstep=32
        integer :: i,j,ntimes,st,l,iupdate
        integer :: inc1,inc2,ld1,ld2,ld3, n, mm,nn,kk

        real(kind) :: R,au,ad,ru,rd,time1,time2,
     &                diagu(1:nt),diagd(1:nt)  

!       encourage unit stride access even in bmd/bmu

        logical, parameter :: use_dgemv = .true.
        logical, parameter :: use_dgemm = .true.
        real(kind) :: alpha_r,beta_r
        real(kind) :: col_j(1:nt), row_j(1:nt),
     &           amu(1:nt,1:maxstep),bmu(1:nt,1:maxstep),
     &           amd(1:nt,1:maxstep),bmd(1:nt,1:maxstep) 
c************************************************************

        time1=real(MPI_Wtime(),kind)
c       initialize the vectors that store the diagonal elements of
c       G needed to calculate M.
        do i=1,nt
          diagu(i) = gu(i,i)
        enddo
        if(isu.eq.1) then
          do i=1,nt
            diagd(i) = gd(i,i)
          enddo
        end if

        do i=1,ntimes ! number of times we sweep through space-time updating
          nslr=nslr+1

          iupdate = 0
          do j=1,nt
c           Now update the j'th ising field element s(j)
            ucount=ucount+oner ! count update attempts
            st=-s(j)
            ru=oner-(diagu(j)-oner)*expm1(+st)
            if(isu.eq.1) then
              rd=oner-(diagd(j)-oner)*expm1(-st)
              R=ru*rd
            else !isu=-1 and U<0 so 
              R=ru**2*exp(-twor*alpha*st)
            end if

c           sprng is David Ceperley's parallel random number generator
c           see http://www.ncsa.uiuc.edu/Apps/CMP/RNG/index_old.html
c           if(abs(R)/(oner+abs(R)).gt.sprng())then   !heatbath
c           if(abs(R)/(oner+abs(R)).gt.rand())then    !heatbath
            if(abs(R).gt.sprng())then                 !metropolis
c             accept the change
              iupdate=iupdate+1
              sgn=sign(oner,R)*sgn
              acrat=acrat+oner ! count accepted changes in s
              s(j)=st

c             store the a and b vectors in am and bm
c             first calculate the new column and row j 
c             from the initial gu and 
c             the matrices am and bm as
c             col_j(i) = gu(i,j) + sum_m am(i,m)*bm(m,j)
c             row_j(i) = gu(j,i) + sum_m am(j,m)*bm(m,i)
c             am(i) = col_j(i)*au - au*delta_i,j
c             bm(i) = row_j(i)

              au=expm1(+st)/ru

              col_j(1:nt) = gu(1:nt,j)
              row_j(1:nt) = gu(j,1:nt)

c             This next loop only is implemented if iupdate>1.
c             It calculates the columns and rows of G, properly
c             accounting for the dependencies.
              if (use_dgemv) then

                  mm = nt
                  nn = iupdate-1
                  alpha_r = oner
                  beta_r = oner

                  ld1 = size(amu,1)
                  inc1 = size(bmu,1)
                  inc2 = 1

                  if (nn.ge.1) then
                  call dgemv( 'N', mm,nn,
     &                alpha_r,amu,ld1,bmu(j,1),inc1,
     &                beta_r,col_j,inc2)
                  endif

                  ld1 = size(bmu,1)
                  inc1 = size(amu,1)
                  inc2 = 1
                  if (nn.ge.1) then
                  call dgemv('N', mm,nn,
     &                alpha_r,bmu,ld1,amu(j,1),inc1,
     &                beta_r,row_j,inc2)
                  endif
              else
                  do n=1,iupdate-1
                     col_j(1:nt) = col_j(1:nt) + amu(1:nt,n)*bmu(j,n)
                     row_j(1:nt) = row_j(1:nt) + amu(j,n)*bmu(1:nt,n)
                  enddo
              endif

              amu(1:nt,iupdate) = au*col_j(1:nt)
              amu(j,iupdate) = amu(j,iupdate) - au

              bmu(1:nt,iupdate) = row_j(1:nt)

c            now update the diagonal
             diagu(1:nt)=diagu(1:nt)+amu(1:nt,iupdate)*bmu(1:nt,iupdate)

             if (isu.eq.1) then ! do the same for the down spin
                ad = expm1(-st)/rd
                col_j(1:nt) = gd(1:nt,j)
                row_j(1:nt) = gd(j,1:nt)

                if (use_dgemv) then

                  mm = nt
                  nn = iupdate-1
                  alpha_r = oner
                  beta_r = oner

                  ld1 = size(amd,1)
                  inc1 = size(bmd,1)
                  inc2 = 1

                  if (nn.ge.1) then
                  call dgemv( 'N', mm,nn,
     &                alpha_r,amd,ld1,bmd(j,1),inc1,
     &                beta_r,col_j,inc2)
                  endif

                  ld1 = size(bmd,1)
                  inc1 = size(amd,1)
                  inc2 = 1
                  if (nn.ge.1) then
                  call dgemv('N', mm,nn,
     &                alpha_r,bmd,ld1,amd(j,1),inc1,
     &                beta_r,row_j,inc2)
                  endif
                else

                do n=1,iupdate-1
                   col_j(1:nt) = col_j(1:nt) + amd(1:nt,n)*bmd(j,n)
                  row_j(1:nt) = row_j(1:nt) + amd(j,n)*bmd(1:nt,n)
                enddo

                endif

                amd(1:nt,iupdate) = ad * col_j(1:nt)
                amd(j,iupdate) = amd(j,iupdate) - ad

                bmd(1:nt,iupdate) = row_j(1:nt)

c            now update the diagonal
             diagd(1:nt)=diagd(1:nt)+amd(1:nt,iupdate)*bmd(1:nt,iupdate)
             endif ! isu.eq.1

c         if iupdate = maxstep we have to update gu and gd according to
c         gu = gu + amu*bmu using dgemm

                  if(iupdate.eq.maxstep) then
                   iupdate=0
                   if (use_dgemm) then
                        mm = nt
                        nn = nt
                        kk = maxstep
                        alpha_r = oner
                        beta_r = oner
                        ld1 = size(amu,1)
                        ld2 = size(bmu,1)
                        ld3 = size(gu,1)

                        call dgemm('N','T',mm,nn,kk,alpha_r,
     &                      amu,ld1, bmu, ld2, beta_r, gu,ld3 )
                        if (isu.eq.1) then
                           ld1 = size(amd,1)
                           ld2 = size(bmd,1)
                           ld3 = size(gd,1)
                           call dgemm('N','T',mm,nn,kk,alpha_r,
     &                          amd,ld1,bmd,ld2,beta_r,gd,ld3)
                         endif
                     else
                       gu(1:nt,1:nt) = gu(1:nt,1:nt) + 
     &                   matmul( amu(1:nt,1:maxstep), 
     &                     transpose( bmu(1:nt,1:maxstep) ) )
                       if (isu.eq.1) then
                         gd(1:nt,1:nt) = gd(1:nt,1:nt) + 
     &                     matmul( amd(1:nt,1:maxstep), 
     &                     transpose( bmd(1:nt,1:maxstep) ) )
                       endif

                    endif ! if (use_dgemm)
                   endif  ! if (iupdate.eq.maxstep)                                             

            end if ! abs(R).gt.sprng()


          end do  ! do j=1,nt

c         Check if the Green function needs updating

                  if(iupdate.gt.0) then
                    if (use_dgemm) then
                        mm = nt
                        nn = nt
                        kk = iupdate
                        alpha_r = oner
                        beta_r = oner
                        ld1 = size(amu,1)
                        ld2 = size(bmu,1)
                        ld3 = size(gu,1)
                        call dgemm('N','T',mm,nn,kk,alpha_r,
     &                      amu,ld1, bmu, ld2, beta_r, gu,ld3 )
                        if (isu.eq.1) then
                           ld1 = size(amd,1)
                           ld2 = size(bmd,1)
                           ld3 = size(gd,1)
                           call dgemm('N','T',mm,nn,kk,alpha_r,
     &                          amd,ld1,bmd,ld2,beta_r,gd,ld3)
                         endif
                     else
                       gu(1:nt,1:nt) = gu(1:nt,1:nt) + 
     &                   matmul( amu(1:nt,1:iupdate), 
     &                     transpose( bmu(1:nt,1:iupdate) ) )
                       if (isu.eq.1) then
                         gd(1:nt,1:nt) = gd(1:nt,1:nt) + 
     &                     matmul( amd(1:nt,1:iupdate), 
     &                     transpose( bmd(1:nt,1:iupdate) ) )
                       endif

                     endif ! if (use_dgemm)

                   iupdate = 0
                   endif                                        
c
c         Now we must check to see if it is time to reconfigure the
c         Gd matrix.  Reconfiguration is necessary to deal with roundoff error.

          if(nslr.ge.ntr)then
            call reg_Ed
            nslr=0
          end if 
        end do

        time2=real(MPI_Wtime(),kind)
        time_date=time_date+time2-time1
        if(myrank.eq.0.and.iprint.ge.9) write(lud,*) 'done update' !0
        return
        end

        subroutine reg_Ed
c       This subroutine recalculates the green functions given only the  
c       s-field configuration and the initial greens functions Gcstsc  
c       (which corresponds to all s=0).

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************

        real(kind) :: uxterm,dxterm,au,ad,ru,rd
        integer i,j,k,l,m,n,ind,imd,ic,jc

!       -----------------------------------------------
!       local arrays related to delayed update by dgemm
!       -----------------------------------------------
        logical, parameter :: use_original = .false.

        logical, parameter :: use_dgemv = .true.
        logical, parameter :: use_dgemm = .true.
        integer, parameter :: maxstep = 32
        integer :: inc1,inc2, ld1,ld2,ld3, mm,nn,kk
        integer :: istart,iend,isize,iupdate
        real(kind) :: alpha_r,beta_r
        real(kind) :: diagu(1:nt), diagd(1:nt)
        real(kind) :: col_i(1:nt),row_i(1:nt)
        real(kind) :: amu(1:nt,1:maxstep),bmu(1:nt,1:maxstep)
        real(kind) :: amd(1:nt,1:maxstep),bmd(1:nt,1:maxstep)


c       Gcstsc is the greens function for all s=0, hence we will
c       sweep through the lattice and reconfigure it to coorespond
c       to the lattice configuration.

c       move Gcstsc to gu and gd
        do m=1,nl
        do n=1,nl
          k=n-m
        do jc=1,Nc
        do ic=1,Nc
          ind=ic + (n-1)*Nc
          imd=jc + (m-1)*Nc
          gu(ind,imd)=Gcst_not(k,ic,jc)
          gd(ind,imd)=Gcst_not(k,ic,jc)
        end do
        end do
        end do
        end do
        sgn=oner 
c       sweep over all the time slices and sites
        if (use_original) then

        do i=1,nt
          uxterm=exp(+alpha*s(i))-oner
          ru=(oner-(gu(i,i)-oner)*uxterm)
          au=uxterm/ru
          call upg_Ed(i,au,gu) 
          if(isu.eq.1) then
c           U>0, so the down spin G must be updated
            dxterm=exp(-alpha*s(i))-oner
            rd=(oner-(gd(i,i)-oner)*dxterm)
            ad=dxterm/rd
            call upg_Ed(i,ad,gd)
            sgn=sign(oner,ru*rd)*sgn
          else ! isu=-1 ( U<0 )
            sgn=oner
          end if
        end do

        else
!        ---------------------
!        use blocking and dgemm
!        ---------------------

         do i=1,nt
           diagu(i) = gu(i,i)
         enddo
         if (isu.eq.1) then
           do i=1,nt
             diagd(i) = gd(i,i)
           enddo
         endif  


         do istart=1,nt,maxstep
          iend = min(nt,istart+maxstep-1)
          isize = iend-istart+1

          do iupdate=1,isize
            i = (istart-1) + iupdate

          uxterm=exp(+alpha*s(i))-oner
          ru=(oner-(diagu(i)-oner)*uxterm)
          au=uxterm/ru

!         --------------
!         setup amu, bmu
!         --------------
          col_i(1:nt) = gu(1:nt,i)
          row_i(1:nt) = gu(i,1:nt)

          if (use_dgemv) then
                  mm = nt
                  nn = iupdate-1
                  alpha_r = oner
                  beta_r = oner

                  ld1 = size(amu,1)
                  inc1 = size(bmu,1)
                  inc2 = 1

                  if (nn.ge.1) then
                  call dgemv( 'N', mm,nn,
     &                alpha_r,amu,ld1,bmu(i,1),inc1,
     &                beta_r,col_i,inc2)
                  endif

                  ld1 = size(bmu,1)
                  inc1 = size(amu,1)
                  inc2 = 1
                  if (nn.ge.1) then
                  call dgemv('N', mm,nn,
     &                alpha_r,bmu,ld1,amu(i,1),inc1,
     &                beta_r,row_i,inc2)
                  endif
          else

          do n=1,iupdate-1
            col_i(1:nt) = col_i(1:nt) + amu(1:nt,n)*bmu(i,n)
            row_i(1:nt) = row_i(1:nt) + amu(i,n)*bmu(1:nt,n)
          enddo

          endif

          amu(1:nt,iupdate) = au*col_i(1:nt)
          amu(i,iupdate) = amu(i,iupdate) - au
          bmu(1:nt,iupdate) = row_i(1:nt)

          diagu(1:nt) = diagu(1:nt) +
     &             amu(1:nt,iupdate)*bmu(1:nt,iupdate)


          if (isu.eq.1) then
c           U>0, so the down spin G must be updated
            dxterm=exp(-alpha*s(i))-oner
            rd=(oner-(diagd(i)-oner)*dxterm)
            ad=dxterm/rd
            sgn=sign(oner,ru*rd)*sgn

!           --------------
!           setup amd, bmd
!           --------------

            col_i(1:nt) = gd(1:nt,i)
            row_i(1:nt) = gd(i,1:nt)

            if (use_dgemv) then 

                  mm = nt
                  nn = iupdate-1
                  alpha_r = oner
                  beta_r = oner

                  ld1 = size(amd,1)
                  inc1 = size(bmd,1)
                  inc2 = 1

                  if (nn.ge.1) then
                  call dgemv( 'N', mm,nn,
     &                alpha_r,amd,ld1,bmd(i,1),inc1,
     &                beta_r,col_i,inc2)
                  endif

                  ld1 = size(bmd,1)
                  inc1 = size(amd,1)
                  inc2 = 1
                  if (nn.ge.1) then
                  call dgemv('N', mm,nn,
     &                alpha_r,bmd,ld1,amd(i,1),inc1,
     &                beta_r,row_i,inc2)
                  endif

            else
            do n=1,iupdate-1
              col_i(1:nt) = col_i(1:nt) + amd(1:nt,n)*bmd(i,n)
              row_i(1:nt) = row_i(1:nt) + amd(i,n)*bmd(1:nt,n)
            enddo
            endif

            amd(1:nt,iupdate) = ad*col_i(1:nt)
            amd(i,iupdate) = amd(i,iupdate) - ad
            bmd(1:nt,iupdate) = row_i(1:nt)

            diagd(1:nt) = diagd(1:nt) + 
     &          amd(1:nt,iupdate) * bmd(1:nt,iupdate)

           else ! isu=-1 ( U<0 )
                   sgn = oner
           endif
           enddo
!          --------------------
!          perform dgemm update
!          --------------------
           if (use_dgemm) then
           mm = nt
           nn = nt
           kk = isize
           alpha_r = oner
           beta_r = oner
           ld1 = size(amu,1)
           ld2 = size(bmu,1)
           ld3 = size(gu,1)
           call dgemm('N','T',mm,nn,kk,
     &       alpha_r,amu,ld1,bmu,ld2,   beta_r,gu,ld3)

           if (isu.eq.1) then
                   ld1 = size(amd,1)
                   ld2 = size(bmd,1)
                   ld3 = size(gd,1)
                   call dgemm('N','T',mm,nn,kk,
     &               alpha_r,amd,ld1, bmd,ld2, beta_r,gd,ld3)
           endif

           else
             gu(1:nt,1:nt) = gu(1:nt,1:nt) + 
     &            matmul(amu(1:nt,1:isize),
     &              transpose(bmu(1:nt,1:isize))  )
             if (isu.eq.1) then
              gd(1:nt,1:nt) = gd(1:nt,1:nt) + 
     &              matmul( amd(1:nt,1:isize),
     &                transpose(bmd(1:nt,1:isize))  )
             endif


            endif ! if (use_dgemm)

           enddo ! do istart

           endif ! if (use_original)

        return
        end

        subroutine upg_Ed(lo,acg,gcg)

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************

        integer lo,i,j
        real(kind) :: acg,d(nt),b(nt),gcg(nt,nt),time1,time2

c       subroutine upg updates gd after a change in one of the field
c       variables has been accepted.

c       FORM THE VECTORS
c       do i=1,nt
c          b(i)=gcg(i,lo)*acg
c          d(i)=gcg(lo,i)
c        end do
        b(:)=gcg(:,lo)*acg
        b(lo)=b(lo)-acg
        d(:)=gcg(lo,:)

c       FORM THE OUTER PRODUCT, AND THE NEW G
c       5 WAYS TO DO IT (see which is fastest on your machine)
cbw     do i=1,nt
cbw     do j=1,nt
cbw       gcg(j,i)=gcg(j,i)+b(j)*d(i)
cbw     end do
cbw     end do
c
c       This is equivalent to the level 2  blas call (usually best)
c
        call sger(nt,nt,oner,b,1,d,1,gcg,nt)
c
c       and or the L3 blas call
c
c       call sgemm('N','T',nt,nt,1,oner,b,nt,d,nt,oner,gcg,nt)
c
c       or to the loop using level 1 blas call
c        
c       do i=1,nt
c         call saxpy(nt,d(i),b,1,gcg(1,i),1)
c       end do
c
c       and to the equivalent fortran90 call
c       do i=1,nt
c         gcg(:,i)=gcg(:,i) +b(:)*d(i)
c       end do

        return
        end
