
        subroutine update_blocked(ntimes)
c*****************************************************************************
c       Update updates the Hirsch-Hubbard-Statonovich fields by sweeping  
c       sequentially through blocks of HHS fields and testing if changes are 
c       likely.  Blocks of Nblock Ising fields are updated by the metropolis 
c       or heat bath algorithm.  If Nblocka changes are accepted, upg updates 
c       the greens functions appropriately. This subroutine should probably be 
c       compiled with an inline option.
c
c       The equations for the old G and new G' green function matrices are
c
c       G' = G + [G - 1][exp(V'-V) - 1]G'                       (1)
c
c       or 
c           -1
c       G G'   = 1 + [1-G][exp(V'-V) - 1]                       (2)
c
c       where here and throughout this discussion, spin lables are 
c       suppressed whenever possible (i.e. Eqs. 1 and 2 apply for Gu and Gd).
c       The matrix [exp(V'-V) - 1] is diagonal with only Nblocka finite 
c       elements, and V=alpha sigma.  As a result, the matrix in Eq. (2) has 
c       a very simple structure
c
c       [ 1 x     x       ]
c       [   X     x       ]
c       [   x 1   x       ]
c       [   x   1 x       ]
c       [   x     X       ]
c       [   x     x 1     ]
c       [   x     x   1   ]
c       [   x     x     1 ]
c     
c       In which only Nblocka columns are finite.  The determinant of such
c       a matrix may be reduced to that of a Nblocka by Nblocka matrix M.
c
c       M_mn = del_mn + [1-G]_mn [exp(V'-V) - 1]_n              (3)
c
c       where n and m are indices of the Nblocka flipped HHS fields.  The
c       transition probability is then set by 
c
c       R = det[Mu] det[Md]                                     (4)
c
c       where u and d are spin labels.  In this case ratio of the probability 
c       of the old to the new configuration, P'/P = R (if U>0).  The 
c       transition probability is then
c
c       P_{s-->s'} = R/(1+R)     heat bath                      (5a)
c       
c       P_{s-->s'} = min(1,R)    Metropolis  (faster)           (5b)
c
c       We accept the change if P_{s-->s'} > random number [0,1].  In this 
c       case, we must update the green function, i.e. calculate G' from G,
c       using Eq. 1.  
c       
c       G'_ij = G_ij + [G - 1]_in [exp(V'-V) - 1]_n G'_nj       (6)
c      
c       Note that  
c       
c       G'_nj = G_nj + [G - 1]_nm [exp(V'-V) - 1]_m G'_mj       (7)
c       
c       or       
c       
c       G'_nj = G_nj - [M_nm-del_mn] G'_mj                      (8)
c      
c       so that 
c                -1 
c       G'_nj = M  _nm G_mj                                     (9)
c       
c       Then, Eq. 6 becomes
c                                                    -1 
c       G'_ij = G_ij + [G - 1]_in [exp(V'-V) - 1]_n M  _nm G_mj        
c
c                                                               (10)
c
c       Thus the green function update takes the form of a nt by Nblocka
c       by Nblocka by nt matrix multiply which may be done with dgemm,
c       provided we form the matrices 
c      
c       B_in = [G - 1]_in 
c                                 -1 
c       D_nj = [exp(V'-V) - 1]_n M  _nm G_mj
c
c                      
c       G'_ij = G_ij + B_in * D_nj        
c
c       first and store them as sequential matrices.  This operation
c       scales like Nblocka times nt; whereas, the matrix multiply
c       in Eq. 10, scales like nt^2 Nblocka.
c       


c*****************************************************************************
        use Global
        implicit none
        include 'mpif.h'
#define SIMPLE_SPRNG
#define FLOAT_GEN
#define USE_MPI
#include "sprng_f.h" 
c*****************************************************************************
        integer :: i,ist,isweep,j,ntimes,l,m,n,Nblocka,indst(Nblock),
     &             indxu(Nblock),indxd(Nblock),st(Nblock),ic,il,
     &             ilstart
        real(kind) :: R,au,ad,ru,rd,time1,time2,Mu(Nblock,Nblock),
     &                Md(Nblock,Nblock),Minv(Nblock,Nblock),du,dd
#ifdef USE_BLOCK_LAPACK
        integer :: info,ncount
        logical :: is_odd
#endif
c*****************************************************************************
 
        time1=real(MPI_Wtime(),kind)
        do isweep=1,ntimes
          nslr=nslr+1
          if(umode.eq.3) then
            Nblocka=max(1,int((Nblock+1)*sprng())) 
c min required to work around "int(0.999999)=1" bug on Cray X1
            Nblocka=min(Nblock,Nblocka) 
          else
            Nblocka=Nblock
          end if
          if(Nblocka.eq.1) then
            ilstart=1
          else
            ilstart=1 + int(nl*sprng())  ! 1<= random integer <= nl
          end if
          do ic=1,Nc
          do il=ilstart,ilstart-1+nl,Nblocka
c           Now update the HHS fields  s(ic,il) to s(ic,il+Nblock). 
            do n=1,Nblocka
              ucount=ucount+oner
              ist=ic + Nc*mod(n-2+il,nl)
              indst(n)=ist
              st(n)=-s(ist)
            end do
            
c
c           M_mn = del_mn + [1-G]_mn [exp(V'-V) - 1]_n  
c
            do n=1,Nblocka
            do m=1,Nblocka
              Mu(m,n)=-gu(indst(m),indst(n))*expm1(+st(n))
            end do
              Mu(n,n)=Mu(n,n)+oner + expm1(+st(n))
            end do

#ifdef USE_BLOCK_LAPACK
            du = oner
            call dgetrf(Nblocka,Nblocka,Mu,Nblock,indxu,info)
            do n=1,Nblocka ! du = det(Mu)
              du=du*Mu(n,n)
            end do
            ncount=0
            do n=1,Nblocka
               if(indxu(n).gt.n) then
                  ncount = ncount+1
               end if
            end do
            is_odd = (mod(ncount,2).eq.1)
            if (is_odd) du =-du
#else
c           calculate the determinant of Mu and Md using numerical recipes
c           calls, Sec. 2.2 (pages 34--41) in the second edition of 
c           numerical recipes with fortran.
            call ludcmp(Mu,Nblocka,Nblock,indxu,du)
            do n=1,Nblocka ! du = det(Mu)
              du=du*Mu(n,n)
            end do
#endif

            if(isu.eq.1) then
              do n=1,Nblocka
              do m=1,Nblocka
                Md(m,n)=-gd(indst(m),indst(n))*expm1(-st(n))
              end do
                Md(n,n)=Md(n,n)+oner + expm1(-st(n))
              end do
#ifdef USE_BLOCK_LAPACK
              dd = oner
              call dgetrf(Nblocka,Nblocka,Md,Nblock,indxd,info)
              do n=1,Nblocka
                dd=dd*Md(n,n)
              end do
              ncount=0
              do n=1,Nblocka
                if(indxd(n).gt.n) then
                        ncount = ncount+1
                end if
              end do
              is_odd = (mod(ncount,2).eq.1)
              if (is_odd) dd=-dd
#else
              call ludcmp(Md,Nblocka,Nblock,indxd,dd)
              do n=1,Nblocka ! dd = det(Md)
                dd=dd*Md(n,n)
              end do
#endif
              R=du*dd
            else !isu=-1 and U<0 so 
              R=du**2*exp(-twor*alpha*sum(st(1:Nblocka)))
            end if

c           sprng is David Ceperley's parallel random number generator
c           see http://www.ncsa.uiuc.edu/Apps/CMP/RNG/index_old.html
c           if(abs(R)/(oner+abs(R)).gt.sprng())then   !heatbath
c           if(abs(R)/(oner+abs(R)).gt.rand())then    !heatbath
            if(abs(R).gt.sprng())then                 !metropolis
c             accept the change
              sgn=sign(oner,R)*sgn
              acrat=acrat+oner*Nblocka
              acount=acount+oner
              do n=1,Nblocka
                s(indst(n))=st(n)
              end do
              call upg_blocked(indst,indxu,Nblocka,Mu,gu,st,expm1,+1)
              if(isu.eq.1) then  ! then Gd must also be updated
                call upg_blocked(indst,indxd,Nblocka,Md,gd,st,expm1,-1)
              end if ! (isu.eq.1)
            end if   ! accept the change
          end do       ! il
          end do       ! ic
c
c         Now we must check to see if it is time to reconfigure the Gd 
c         matrix.  Reconfiguration is necessary to deal with roundoff error.
 
          if(nslr.ge.ntr)then
            call reg_blocked
            nslr=0
          end if 
        end do  ! isweep
 
        time2=real(MPI_Wtime(),kind)
        time_date=time_date+time2-time1
        if(myrank.eq.0.and.iprint.ge.9) write(lud,*) 'done update' !0
        return
        end

c*******************************************************************************
        subroutine reg_blocked
c*******************************************************************************
c       This subroutine recalculates the green functions given only the  
c       s-field configuration and the initial green's functions Gcstsc  
c       (which corresponds to all s=0).
c*******************************************************************************
        use Global
        implicit none
        include 'mpif.h'
c*******************************************************************************
 
        integer :: i,ist,isweep,ic,ind,imd,jc,j,k,l,m,n,Nblocka,
     &             indst(Nblock),indxu(Nblock),indxd(Nblock),st(Nblock)
        real(kind) :: Mu(Nblock,Nblock),Md(Nblock,Nblock),du,dd,
     &                Minv(Nblock,Nblock)
#ifdef USE_BLOCK_LAPACK
        integer :: info,ncount
        logical :: is_odd
#endif
 
c*******************************************************************************
c       Gcstsc is the greens function for all s=0, hence we will
c       sweep through the lattice and reconfigure it to coorespond
c       to the lattice configuration.
 
c       move Gcstsc to gu and gd
        do m=1,nl
        do n=1,nl
          k=n-m
        do ic=1,Nc
        do jc=1,Nc
          ind=ic + (n-1)*Nc
          imd=jc + (m-1)*Nc
          gu(ind,imd)=Gcst_not(k,ic,jc)
          gd(ind,imd)=Gcst_not(k,ic,jc)
        end do
        end do
        end do
        end do
        sgn=oner 
c       sweep over all the time slices and sites
        do ist=1,nt,Nblock
          Nblocka=0
          do n=ist,min(ist+Nblock-1,nt)
            Nblocka=Nblocka+1
            indst(Nblocka)=n
          end do

c	   copy the spins
c ct : stores a piece of the ising field (Only the spins that are
c  touched by the block. 
          st(1:Nblocka)=s(ist:ist+Nblocka-1)

          do n=1,Nblocka
            do m=1,Nblocka
            Mu(m,n)=-gu(indst(m),indst(n))*expregm1(+st(n))
            end do

            Mu(n,n)=Mu(n,n) + oner + expregm1(+st(n))
          end do

#ifdef USE_BLOCK_LAPACK
          du = oner
          call dgetrf(Nblocka,Nblocka,Mu,Nblock,indxu,info)
          do n=1,Nblocka
            du=du*Mu(n,n)
          end do
          ncount=0
          do n=1,Nblocka
             if(indxu(n).gt.n) then
                ncount = ncount+1
             end if
          end do
          is_odd = (mod(ncount,2).eq.1)
          if (is_odd) du=-du
#else
          call ludcmp(Mu,Nblocka,Nblock,indxu,du)
          do n=1,Nblocka ! du = det(Mu)
            du=du*Mu(n,n)
          end do
#endif
          call upg_blocked(indst,indxu,Nblocka,Mu,gu,st,expregm1,+1)
          if(isu.eq.1) then
c           U>0, so the down spin G must be updated
            do n=1,Nblocka
            do m=1,Nblocka
              Md(m,n)=-gd(indst(m),indst(n))*expregm1(-st(n))
            end do
              Md(n,n)=Md(n,n)+oner + expregm1(-st(n))
            end do
#ifdef USE_BLOCK_LAPACK
            dd = oner
            call dgetrf(Nblocka,Nblocka,Md,Nblock,indxd,info)
            do n=1,Nblocka
               dd=dd*Md(n,n)
            end do
            ncount=0
            do n=1,Nblocka
               if(indxd(n).gt.n) then
                  ncount = ncount+1
               end if
            end do
            is_odd = (mod(ncount,2).eq.1)
            if (is_odd) dd=-dd
#else
            call ludcmp(Md,Nblocka,Nblock,indxd,dd)
            do n=1,Nblocka
              dd=dd*Md(n,n)
            end do
#endif
            call upg_blocked(indst,indxd,Nblocka,Md,gd,st,expregm1,-1)
            sgn=sign(oner,du*dd)*sgn
          else ! isu=-1 ( U<0 )
            sgn=oner
          end if
        end do
 
        return
        end

c*******************************************************************************
        subroutine upg_blocked(indst,indx,Nblocka, 
     &     Mcg,gcg,st,expm1cg,iud)
c*******************************************************************************
        use Global
        implicit none
        include 'mpif.h'
c*******************************************************************************
 
        integer :: i,iud,j,n,m,Nblocka,indst(Nblock),indx(Nblock),
     &             st(Nblock)
        real(kind) ::gcg(nt,nt),Mcg(Nblock,Nblock),Minv(Nblock,Nblock),
     &               expm1cg(-1:1)
#ifdef USE_BLOCK_LAPACK
        integer :: info
        real(kind) :: work(Nblock*64)
#endif

 
c*******************************************************************************

c       
c       We need to form
c                                 -1 
c       D_nj = [exp(V'-V) - 1]_n M  _nm G_mj
c
#ifdef USE_BLOCK_LAPACK
c       First, form the inverse of M.
        call dgetri(Nblocka,Mcg,Nblock,indx,work,Nblock*64,info)
c                             -1
c       Minv now contains  Mcg  
c
        do m=1,Nblocka
        do n=1,Nblocka
          Mcg(n,m)=expm1cg(iud*st(n))*Mcg(n,m)
        end do
        end do
#else
c       First, form the inverse of M.
        Minv=zeror
        do n=1,Nblocka
          Minv(n,n)=oner
        end do
        do n=1,Nblocka
          call lubksb(Mcg,Nblocka,Nblock,indx,Minv(1,n))
        end do
c                             -1
c       Minv now contains  Mcg  
c
        do m=1,Nblocka
        do n=1,Nblocka
          Mcg(n,m)=expm1cg(iud*st(n))*Minv(n,m)
        end do
        end do
#endif
c                                             -1 
c       Mcg now contains = [exp(V'-V) - 1]_n M  _nm 
c

c       find the inverse of Mu and make the matrices needed to
c       update G.
c       
c       B_in = [G - 1]_in 
c                                 
        do i=1,nt
        do n=1,Nblocka
          Bmat(i,n) = gcg(i,indst(n))
        end do
        end do
        do n=1,Nblocka
          i=indst(n)
          Bmat(i,n)=Bmat(i,n)-oner
        end do

c       Now form D
        Dmat=zeror
        do j=1,nt
        do n=1,Nblocka
        do m=1,Nblocka
          Dmat(n,j)=Dmat(n,j)+Mcg(n,m)*gcg(indst(m),j)
        end do
        end do
        end do
        
c       now use dgmemm to update gu
c                
c       G'_ij = G_ij + B_in * D_nj        
c
c       SGEMM  - perform one of the matrix-matrix operations   
c       C := alpha*op( A )*op( B ) + beta*C,
c       SUBROUTINE SGEMM ( TRANSA, TRANSB, M, N, K,  ALPHA,  A,  LDA,  
c                          B,  LDB, BETA, C, LDC )
        
        call dgemm('N','N',nt,nt,Nblocka,oner,Bmat,nt,Dmat,Nblock,
     &             oner,gcg,nt)



        return
        end

c*******************************************************************************
c*******************************************************************************
c*******************************************************************************

