        subroutine update_delayed(ntimes)
c*****************************************************************************
c       Update updates the Hubbard-Statonovich fields by sweeping sequentially 
c       through them and testing if changes are required.  The Ising field is
c       updated by the Metropolis or heat bath algorithm.  If changes are made 
c       it updates the Green's functions appropriately.  This block should 
c       probably be compiled with an inline option.
c       
c       This version is based on Eduardo DAzevedos proposed changes 
c       to delay the dger updates and perform a slender rectangular dgemm
c       update.  This method is numerically identical to the original 
c       Hirsch-Fye algorithm i.e. it proposes the same changes with the 
c       same acceptance probability; however, by delaying the Green function 
c       update, it circumvents the associated caching problems. The HF 
c       algorithm sweeps sequentially through blocks of HHS fields and 
c       proposes local changes in the HHS fields
c
c       s_n --> s_n' = -s_n
c
c       The transition probability is calculated from 
c
c               -1
c       det[G G'   ] = det [I + [I-G][exp(V'-V) - I]  ]                    
c
c       The diagonal matrix [exp(V'-V) - 1] has only 1 finite element, where 
c       V=alpha sigma s_n.  So
c
c                    -1
c       R = det[G G'   ] = I + [1-G_nn][exp(V_n'-V_n) - 1] = M    (1)                
c
c
c       If the change is accepted, we update the Green functions using 
c
c
c       G' = G + [G - 1][exp(V'-V) - 1]G'                       (2)
c
c       We iterate Eq. 2 twice to obtain an updating equation for G
c
c                                                     
c       G'_ij = G_ij + [G - 1]_in [exp(V'-V) - 1]_n/M   G_nj    (3)    
c
c       This first equation is a vector outer product followed by a
c       matrix update, and may be completed with a call to dger.
c       
c       However, dger is very innefficient, since it has only two 
c       floating point operations per memory access.  Both FPOs
c       can be done on the same clock cycle, but many clock cycles are
c       required for the memory access.  Thus Eq. 3 is limited 
c       by the speed of the memory bus, how fast G can be moved in
c       and out of memory.
c
c       Ed proposes that we delay the Green function update, until a 
c       sufficiently large number of changes have been accumulated.  
c       In this example we set the number of accepted changes, Nblock=32.
c       Only after 32 changes in the HHS fields is the whole Green function 
c       matrix updated.  If we label these 32 accepted changes sequentially, 
c       then Eq. 3 becomes 
c
c        32      1        31                              31    31
c       G_ ij = G_ ij + [G - 1]_in31 [exp(V'-V) - 1]_n31/M     G_ n31j     
c
c                 30                              30    30
c             + [G - 1]_in30 [exp(V'-V) - 1]_n30/M     G_ n30j     
c             .
c             .
c             .
c                 1                             1     1
c             + [G - 1]_in1 [exp(V'-V) - 1]_n1/M     G_ n1j          (4)
c
c       where n1, n2, n3... are the indices of the first, second, third...
c       accepted changed field.  Again, this may be written as a series
c       of vector outer products.  If we identify
c
c        N      N                             N 
c       A_i = [G - 1]_inN [exp(V'-V) - 1]_nN/M                      (5a)
c
c        N     N
c       B_j = G_nNj                                                 (5b)
c
c       then
c        N      1     N-1  n   n
c       G_ij = G_ij + SUM A_i B_j                                   (6a)
c                     n=1
c       This can be collaped into a matrix matrix product (nt by 32
c       matrix A and 32 by nt matrix B)
c
c        32   1
c       G  = G  + A * B                                             (6b)
c
c       Since each A^n and B^n depends on the prior G, this step is pointless
c       unless some way may be found to calculate A^n and B^n for each n which
c       scales like O(nt).  Ed noted that it only takes of O(32*nt) operations 
c       to account for these dependencies since Eqs. 5 don't require the 
c       entire G, but rather only one column and row of G.
c
c        N         N                                    N -1
c       A_i = [colnN_i - delta_ij] [exp(V'-V) - 1]_nN M             (7a)
c
c        N        N
c       B_j = rown_j                                                (7b)
c
c           N        1    N-1  n   n
c       coln_i = coln_i + SUM A_i B_n                               (7c)
c                         n=1
c
c           N        1    N-1  n   n
c       rown_j = rown_j + SUM A_n B_j                               (7d)
c                         n=1
c
c       Furthermore, the row and column information is needed only to 
c       calculate the corressponding A^n and B^n, and may be discarded 
c       afterwards.  A similar equation is used to calculate the diagonal
c       elements of G needed for the calculation of M.
c
c            N     1     N-1  n   n
c       diagn_i = G_ii + SUM A_i B_i                                 (7e)
c                        n=1
c
c       Together, Eqs 6-7 form a very efficient Green function update.  
c       Eqs. 7a and b only require O(nt) operations, Eqs. 7c and d require 
c       only O(nt*N) operations. Eq. 6a or 6b are the only ones in the set 
c       with O(nt^2) operations, but they have many FLOPs per memory access.
c
c       Note that Nblock must not be too large.  As Nblock approaches nt, 
c       Eqs. 7 c and d will begin to scale like O(nt^2).  I.e. the amount of 
c       cpu time required to track the dependences noted above will eventually 
c       consume the same amount of work required to update G.  So, 
c       as Nblock-->nt, this process is likely to become less efficient.   
c       Fortunately, an 1 << Nblock <<nt is sufficient to overcome the caching 
c       problem in the green function update without approaching nt.  
c       Furthermore, Eqs. 7c and d will always have many more FLOP than
c       memory accesses, so this algorithm is likely to always be superior
c       to the original one of Hirsch and Fye.
c
c
c*****************************************************************************
        use Global
        use pTiming
        implicit none
        include 'mpif.h'
#define SIMPLE_SPRNG
#define FLOAT_GEN
#define USE_MPI
#include "sprng_f.h" 
c************************************************************
c        integer, parameter :: maxstep=32
        integer :: i,j,ntimes,st,l,iupdate
        integer :: inc1,inc2,ld1,ld2,ld3, n, mm,nn,kk

        real(kind) :: R,au,ad,ru,rd,time1,time2,
     &                diagu(1:nt),diagd(1:nt)  

!       encourage unit stride access even in bmd/bmu

        logical, parameter :: use_dgemv = .false.
        logical, parameter :: use_dgemm = .true.
        real(kind) :: alpha_r,beta_r
        real(kind) :: col_j(1:nt), row_j(1:nt),
     &           amu(1:nt,1:Nblock),bmu(1:nt,1:Nblock),
     &           amd(1:nt,1:Nblock),bmd(1:nt,1:Nblock) 
c************************************************************

        accumlated_building = 0.0
	accumlated_mul = 0.0
         write (*,'(A,A,A,A,A)') "Delay",tab,"A,B build",tab, "dgemm"
c       start_time=real(MPI_Wtime(),kind)
c       initialize the vectors that store the diagonal elements of
c       G needed to calculate M.
        do i=1,nt
          diagu(i) = gu(i,i)
        enddo
        if(isu.eq.1) then
          do i=1,nt
            diagd(i) = gd(i,i)
          enddo
        end if

        do i=1,ntimes ! number of times we sweep through space-time updating
          nslr=nslr+1

          iupdate = 0
          do j=1,nt 

            start_time=real(MPI_Wtime(),kind)
c           Now update the j'th ising field element s(j)
            ucount=ucount+oner ! count update attempts
c
c                        -1
c           R = det[G G'   ] = I + [1-G_nn][exp(V_n'-V_n) - 1] = M    (1)                
c
            st=-s(j)
            ru=oner-(diagu(j)-oner)*expm1(+st)
            if(isu.eq.1) then
              rd=oner-(diagd(j)-oner)*expm1(-st)
              R=ru*rd
            else !isu=-1 and U<0 so 
              R=ru**2*exp(-twor*alpha*st)
            end if
c           sprng is David Ceperley's parallel random number generator
c           see http://www.ncsa.uiuc.edu/Apps/CMP/RNG/index_old.html
c           if(abs(R)/(oner+abs(R)).gt.sprng())then   !heatbath
c           if(abs(R)/(oner+abs(R)).gt.rand())then    !heatbath
            if(abs(R).gt.sprng())then                 !metropolis
c             accept the change
              iupdate=iupdate+1
              sgn=sign(oner,R)*sgn
              acrat=acrat+oner ! count accepted changes in s
              s(j)=st

c             Store the A and B vectors in Am and Bm
c             first calculate the new column and row j 
c             from the initial gu and 
c             the matrices am and bm as
c
c                 N        1    N-1  n   n
c             coln_i = coln_i + SUM A_i B_n                          (7c)
c                               n=1
c
c                 N        1    N-1  n   n
c             rown_j = rown_j + SUM A_n B_j                          (7d)
c                               n=1
c

              au=expm1(+st)/ru
              col_j(1:nt) = gu(1:nt,j)
              row_j(1:nt) = gu(j,1:nt)

c             This next loop only is implemented if iupdate>1.
c             It calculates the columns and rows of G, properly
c             accounting for the dependencies.
              if (use_dgemv) then
c                 Eqs. 7c and 7d
                  mm = nt
                  nn = iupdate-1
                  alpha_r = oner
                  beta_r = oner

                  ld1 = size(amu,1)
                  inc1 = size(bmu,1)
                  inc2 = 1

                  if (nn.ge.1) then
                  call dgemv( 'N', mm,nn,
     &                alpha_r,amu,ld1,bmu(j,1),inc1,
     &                beta_r,col_j,inc2)
                  endif

                  ld1 = size(bmu,1)
                  inc1 = size(amu,1)
                  inc2 = 1
                  if (nn.ge.1) then
                  call dgemv('N', mm,nn,
     &                alpha_r,bmu,ld1,amu(j,1),inc1,
     &                beta_r,row_j,inc2)
                  endif
              else
                  do n=1,iupdate-1
                     col_j(1:nt) = col_j(1:nt) + amu(1:nt,n)*bmu(j,n)
                     row_j(1:nt) = row_j(1:nt) + amu(j,n)*bmu(1:nt,n)
                  enddo
              endif

c                  
c
c              N         N                                    N -1
c             A_i = [colnN_i - delta_ij] [exp(V'-V) - 1]_nN M          (7a)
c
c              N        N
c             B_j = rown_j                                             (7b)
c
              amu(1:nt,iupdate) = au*col_j(1:nt)
              amu(j,iupdate) = amu(j,iupdate) - au
              bmu(1:nt,iupdate) = row_j(1:nt)

c            Now update the diagonal
c
c                 N     1     N-1  n   n
c            diagn_i = G_ii + SUM A_i B_i                              (7e)
c                             n=1
c
             diagu(1:nt)=diagu(1:nt)+amu(1:nt,iupdate)*bmu(1:nt,iupdate)

             if (isu.eq.1) then ! do the same for the down spin
c               Store the A and B vectors in Am and Bm
c               first calculate the new column and row j 
c               from the initial gu and 
c               the matrices am and bm as
c
c                   N        1    N-1  n   n
c               coln_i = coln_i + SUM A_i B_n                          (7c)
c                                 n=1
c
c                   N        1    N-1  n   n
c               rown_j = rown_j + SUM A_n B_j                          (7d)
c                                 n=1
c
                ad = expm1(-st)/rd
                col_j(1:nt) = gd(1:nt,j)
                row_j(1:nt) = gd(j,1:nt)

                if (use_dgemv) then

                  mm = nt
                  nn = iupdate-1
                  alpha_r = oner
                  beta_r = oner

                  ld1 = size(amd,1)
                  inc1 = size(bmd,1)
                  inc2 = 1

                  if (nn.ge.1) then
                  call dgemv( 'N', mm,nn,
     &                alpha_r,amd,ld1,bmd(j,1),inc1,
     &                beta_r,col_j,inc2)
                  endif

                  ld1 = size(bmd,1)
                  inc1 = size(amd,1)
                  inc2 = 1
                  if (nn.ge.1) then
                  call dgemv('N', mm,nn,
     &                alpha_r,bmd,ld1,amd(j,1),inc1,
     &                beta_r,row_j,inc2)
                  endif
                else
                  do n=1,iupdate-1
                     col_j(1:nt) = col_j(1:nt) + amd(1:nt,n)*bmd(j,n)
                     row_j(1:nt) = row_j(1:nt) + amd(j,n)*bmd(1:nt,n)
                  enddo
                endif

c                    
c
c                N         N                                    N -1
c               A_i = [colnN_i - delta_ij] [exp(V'-V) - 1]_nN M          (7a)
c
c                N        N
c               B_j = rown_j                                             (7b)
c
                amd(1:nt,iupdate) = ad * col_j(1:nt)
                amd(j,iupdate) = amd(j,iupdate) - ad

                bmd(1:nt,iupdate) = row_j(1:nt)

c               now update the diagonal
c
c                    N     1     N-1  n   n
c               diagn_i = G_ii + SUM A_i B_i                            (7e)
c                                n=1
c
                diagd(1:nt)=diagd(1:nt)+amd(1:nt,iupdate)*bmd(1:nt,iupdate)
             endif ! isu.eq.1

c            If iupdate = Nblock we have to update gu and gd according to
c
c              N      1     N-1  n   n
c             G_ij = G_ij + SUM A_i B_j                   (6a)
c                           n=1
c


             if(iupdate.eq.Nblock) then
               write (*,'(I)',advance="no") iupdate
               end_time=real(MPI_Wtime(),kind)
               write (*,'(F,A)', advance="no") end_time - start_time,tab
               accumlated_building = accumlated_building + end_time - start_time

               iupdate=0
               start_time=real(MPI_Wtime(),kind)
               if (use_dgemm) then
                    mm = nt
                    nn = nt
                    kk = Nblock
                    alpha_r = oner
                    beta_r = oner
                    ld1 = size(amu,1)
                    ld2 = size(bmu,1)
                    ld3 = size(gu,1)

                    call dgemm('N','T',mm,nn,kk,alpha_r,
     &                  amu,ld1, bmu, ld2, beta_r, gu,ld3 )
                    if (isu.eq.1) then
                       ld1 = size(amd,1)
                       ld2 = size(bmd,1)
                       ld3 = size(gd,1)
                       call dgemm('N','T',mm,nn,kk,alpha_r,
     &                      amd,ld1,bmd,ld2,beta_r,gd,ld3)
                     endif
               else
                     gu(1:nt,1:nt) = gu(1:nt,1:nt) + 
     &                             matmul( amu(1:nt,1:Nblock), 
     &                             transpose( bmu(1:nt,1:Nblock) ) )
                     if (isu.eq.1) then
                       gd(1:nt,1:nt) = gd(1:nt,1:nt) + 
     &                                 matmul( amd(1:nt,1:Nblock), 
     &                                 transpose( bmd(1:nt,1:Nblock) ) )
                     endif
               endif ! if (use_dgemm)
             end_time=real(MPI_Wtime(),kind)
             accumlated_mul = accumlated_mul + end_time - start_time
             write (*,'(F,A)') end_time - start_time,tab
             endif  ! if (iupdate.eq.Nblock)/                    

            end if ! abs(R).gt.sprng()

          end do  ! do j=1,nt

c         Check if the Green function needs updating (i.e. when nt is not
c         divisible by Nblock)

          if(iupdate.gt.0) then
       write (*,'(I)',advance="no") iupdate
       end_time=real(MPI_Wtime(),kind)
       write (*,'(F,A)', advance="no") end_time - start_time,tab
           start_time=real(MPI_Wtime(),kind)
            if (use_dgemm) then
               mm = nt
               nn = nt
               kk = iupdate
               alpha_r = oner
               beta_r = oner
               ld1 = size(amu,1)
               ld2 = size(bmu,1)
               ld3 = size(gu,1)
               call dgemm('N','T',mm,nn,kk,alpha_r,
     &    	   amu,ld1, bmu, ld2, beta_r, gu,ld3 )
               if (isu.eq.1) then
          	  ld1 = size(amd,1)
          	  ld2 = size(bmd,1)
          	  ld3 = size(gd,1)
          	  call dgemm('N','T',mm,nn,kk,alpha_r,
     &    	       amd,ld1,bmd,ld2,beta_r,gd,ld3)
          	endif
            else
              gu(1:nt,1:nt) = gu(1:nt,1:nt) + 
     &    	              matmul( amu(1:nt,1:iupdate), 
     &    	                     transpose( bmu(1:nt,1:iupdate) ) )
              if (isu.eq.1) then
          	gd(1:nt,1:nt) = gd(1:nt,1:nt) + 
     &    	                matmul( amd(1:nt,1:iupdate), 
     &    	                transpose( bmd(1:nt,1:iupdate) ) )
              endif

            endif ! if (use_dgemm)
            end_time=real(MPI_Wtime(),kind)
            accumlated_mul = accumlated_mul + end_time - start_time
C            write (*,*) "We have As and Bs that dont count to a whole A,B"
	    write (*,'(F,A)') end_time - start_time,tab

            iupdate = 0
          endif ! if(iupdate.gt.0)				       
c
c         Now we must check to see if it is time to reconfigure the
c         Gd matrix.  Reconfiguration is necessary to deal with roundoff error.

          if(nslr.ge.ntr)then
	    write (*,*) "Call reg_delayed from update_delayed"
            call reg_delayed
            nslr=0
          end if 
        end do ! do i=1,ntimes

	write (*,*) "update_delayed"
        write (*,'(A,A,A,A,A,A,A,A)')tab,"Delays",tab,"A,B_builds",tab, "dgemms",tab, "All"
        write (*,'(I,F,F,F)')  Nblock, accumlated_building, accumlated_mul, accumlated_mul + accumlated_building
        if(myrank.eq.0.and.iprint.ge.9) write(lud,*) 'done update' !0
        return
        end

        subroutine reg_delayed
c       This subroutine recalculates the green functions given only the  
c       s-field configuration and the initial greens functions Gcstsc  
c       (which corresponds to all s=0).

c************************************************************
        use Global
        use pTiming
        implicit none
        include 'mpif.h'
c************************************************************

        real(kind) :: uxterm,dxterm,au,ad,ru,rd
        integer i,j,k,l,m,n,ind,imd,ic,jc

!       -----------------------------------------------
!       local arrays related to delayed update by dgemm
!       -----------------------------------------------

        logical, parameter :: use_dgemv = .false.
        logical, parameter :: use_dgemm = .true.
c        integer, parameter :: maxstep = 32
        integer :: inc1,inc2, ld1,ld2,ld3, mm,nn,kk
        integer :: istart,iend,isize,iupdate
        real(kind) :: alpha_r,beta_r
        real(kind) :: diagu(1:nt), diagd(1:nt)
        real(kind) :: col_i(1:nt),row_i(1:nt)
        real(kind) :: amu(1:nt,1:Nblock),bmu(1:nt,1:Nblock)
        real(kind) :: amd(1:nt,1:Nblock),bmd(1:nt,1:Nblock)
        accumlated_building = 0.0
	accumlated_mul = 0.0
c       Gcstsc is the greens function for all s=0, hence we will
c       sweep through the lattice and reconfigure it to coorespond
c       to the lattice configuration.

c       move Gcstsc to gu and gd
        do m=1,nl
        do n=1,nl
          k=n-m
        do jc=1,Nc
        do ic=1,Nc
          ind=ic + (n-1)*Nc
          imd=jc + (m-1)*Nc
          gu(ind,imd)=Gcst_not(k,ic,jc)
          gd(ind,imd)=Gcst_not(k,ic,jc)
        end do
        end do
        end do
        end do
        sgn=oner 
c       sweep over all the time slices and sites
!        ---------------------
!        use blocking and dgemm
!        ---------------------

         do i=1,nt
           diagu(i) = gu(i,i)
         enddo
         if (isu.eq.1) then
           do i=1,nt
             diagd(i) = gd(i,i)
           enddo
         endif  

         write (*,'(A,A,A,A,A)') "Delay",tab,"A,B build",tab, "dgemm"


         do istart=1,nt,Nblock
          iend = min(nt,istart+Nblock-1)
          isize = iend-istart+1

          write (*,'(I,a)',advance="no") istart/Nblock,tab
          start_time = MPI_WTIME() !pTime

          do iupdate=1,isize
            i = (istart-1) + iupdate

          uxterm=exp(+alpha*s(i))-oner
          ru=(oner-(diagu(i)-oner)*uxterm)
          au=uxterm/ru

!         --------------
!         setup amu, bmu
!         --------------
          col_i(1:nt) = gu(1:nt,i)
          row_i(1:nt) = gu(i,1:nt)

          if (use_dgemv) then
                  mm = nt
                  nn = iupdate-1
                  alpha_r = oner
                  beta_r = oner

                  ld1 = size(amu,1)
                  inc1 = size(bmu,1)
                  inc2 = 1

                  if (nn.ge.1) then
                  call dgemv( 'N', mm,nn,
     &                alpha_r,amu,ld1,bmu(i,1),inc1,
     &                beta_r,col_i,inc2)
                  endif

                  ld1 = size(bmu,1)
                  inc1 = size(amu,1)
                  inc2 = 1
                  if (nn.ge.1) then
                  call dgemv('N', mm,nn,
     &                alpha_r,bmu,ld1,amu(i,1),inc1,
     &                beta_r,row_i,inc2)
                  endif
          else

          do n=1,iupdate-1
            col_i(1:nt) = col_i(1:nt) + amu(1:nt,n)*bmu(i,n)
            row_i(1:nt) = row_i(1:nt) + amu(i,n)*bmu(1:nt,n)
          enddo

          endif

          amu(1:nt,iupdate) = au*col_i(1:nt)
          amu(i,iupdate) = amu(i,iupdate) - au
          bmu(1:nt,iupdate) = row_i(1:nt)

          diagu(1:nt) = diagu(1:nt) +
     &             amu(1:nt,iupdate)*bmu(1:nt,iupdate)


          if (isu.eq.1) then
c           U>0, so the down spin G must be updated
            dxterm=exp(-alpha*s(i))-oner
            rd=(oner-(diagd(i)-oner)*dxterm)
            ad=dxterm/rd
            sgn=sign(oner,ru*rd)*sgn

!           --------------
!           setup amd, bmd
!           --------------

            col_i(1:nt) = gd(1:nt,i)
            row_i(1:nt) = gd(i,1:nt)

            if (use_dgemv) then 

                  mm = nt
                  nn = iupdate-1
                  alpha_r = oner
                  beta_r = oner

                  ld1 = size(amd,1)
                  inc1 = size(bmd,1)
                  inc2 = 1

                  if (nn.ge.1) then
                  call dgemv( 'N', mm,nn,
     &                alpha_r,amd,ld1,bmd(i,1),inc1,
     &                beta_r,col_i,inc2)
                  endif

                  ld1 = size(bmd,1)
                  inc1 = size(amd,1)
                  inc2 = 1
                  if (nn.ge.1) then
                  call dgemv('N', mm,nn,
     &                alpha_r,bmd,ld1,amd(i,1),inc1,
     &                beta_r,row_i,inc2)
                  endif

            else
            do n=1,iupdate-1
              col_i(1:nt) = col_i(1:nt) + amd(1:nt,n)*bmd(i,n)
              row_i(1:nt) = row_i(1:nt) + amd(i,n)*bmd(1:nt,n)
            enddo
            endif

            amd(1:nt,iupdate) = ad*col_i(1:nt)
            amd(i,iupdate) = amd(i,iupdate) - ad
            bmd(1:nt,iupdate) = row_i(1:nt)

            diagd(1:nt) = diagd(1:nt) + 
     &          amd(1:nt,iupdate) * bmd(1:nt,iupdate)

           else ! isu=-1 ( U<0 )
                   sgn = oner
           endif
           enddo
           end_time = MPI_WTIME()
           write (*,'(F,A)',advance="no") end_time - start_time,tab
           accumlated_building = accumlated_building + end_time - start_time
!          --------------------
!          perform dgemm update
!          --------------------
                start_time = MPI_WTIME()
           if (use_dgemm) then
           mm = nt
           nn = nt
           kk = isize
           alpha_r = oner
           beta_r = oner
           ld1 = size(amu,1)
           ld2 = size(bmu,1)
           ld3 = size(gu,1)
           call dgemm('N','T',mm,nn,kk,
     &       alpha_r,amu,ld1,bmu,ld2,   beta_r,gu,ld3)

           if (isu.eq.1) then
                   ld1 = size(amd,1)
                   ld2 = size(bmd,1)
                   ld3 = size(gd,1)
                   call dgemm('N','T',mm,nn,kk,
     &               alpha_r,amd,ld1, bmd,ld2, beta_r,gd,ld3)
           endif

           else
             gu(1:nt,1:nt) = gu(1:nt,1:nt) + 
     &            matmul(amu(1:nt,1:isize),
     &              transpose(bmu(1:nt,1:isize))  )
             if (isu.eq.1) then
              gd(1:nt,1:nt) = gd(1:nt,1:nt) + 
     &              matmul( amd(1:nt,1:isize),
     &                transpose(bmd(1:nt,1:isize))  )
             endif


            endif ! if (use_dgemm)

        end_time = MPI_WTIME()
        write (*,'(F,A)') end_time - start_time,tab
        accumlated_mul = accumlated_mul + end_time - start_time
           enddo ! do istart

	write (*,*) "reg_delayed"
        write (*,'(A,A,A,A,A,A,A,A)') tab,"Delays",tab,"A,B_builds",tab, "dgemms",tab, "All"
        write (*,'(I,F,F,F)')  Nblock, accumlated_building, accumlated_mul, accumlated_mul + accumlated_building

        return
        end
