c       This block contains the model-specific parts of the DCA code for 
c       the Hubbard model.
c************************************************************************** 
c       This program simulates the 2-dimensional hubbard model in the
c       dynamical cluster approximation
c
c                        H=H1+H2
c                             +
c                H1= -t SUM ((C (i,s)C(j,s)) + h.c. )
c                       <ij>s
c                               +  
c                    -t' SUM ((C (i,s)C(j,s)) + h.c. )
c                       [ij]s                           i,j n.n.n.
c                                         +
c                        SUM ((t_perp(k) C (k,i,s)C(k,j,s)) + h.c. )
c                        (i,j)s                         i,j adjacent planes
c
c                H2= U SUM (ndu-1/2)(ndd-1/2) + (V + ed)*( ndu + ndd )
c                       i     i        i          i           i     i
c
c       where V_i is the diagonal disorder potential and can have a 
c       continuous (V0<0) or a binary (V0>0) distribution, and
c       tperp(k) = -2 tperp (cos(kx) - cos(ky))^2 for a bilayer Cu02
c       or tperp(k) = -2 tperp for a simple n.n. interlayer coupling.
c
c       H is particle-hole symmetric whenever ed=0 and t'=0.  For a 
c       two-dimensional system, this means that 
c
c               +             +    i
c       C  --> C exp(iQ.i) = C (-1)
c        i      i             i
c                                                +              +
c       C  = SUM exp(ik.i) C  --> SUM exp(ik.i) C exp(-iQ.i) = C
c        k    i             i                    i              Q-k
c
c       or if ed=0, G (tau) = -G   (-tau) = G   (beta-tau)
c                    k          Q-k          Q-k
c       and
c                   G (iwn) = - G   (-iwn)  real parts opposite sign
c                    k           Q-k        imaginary parts equal sign
c
c       The algorithm is three dimensional: the imaginary time
c       direction and the two spatial directions.  The interplanar
c       coupling is treated only at the (dynamical) mean field level.
c       The imaginary time is indexed so the l=1, corresponds to
c       tau=0.0, and l=nl corresponds to tau=beta-dt.
c
c**************************************************************************
c       DEVELOPERS
c       Original serial code by Mark Jarrell and Niki Tahvildarzadeh, 1998.
c
c       This is a perfectly parallel version was written by
c       Carey Huscroft 1998 and Mark Jarrell 1999-2000.  
c
c       Modified by M. Jarrell, A. Macridin and Th. Maier to treat
c       odd cluster sizes (Nc=10,20,26...), late 2003.
c
c       Modified by Th. Maier to treat quasi-two dimensional lattices
c       (ndim=3), and by M. Jarrell and Th. Pruschke to treat fully 
c       three dimensional lattices using Betts cluster in 2004.  Modified
c       to use 2D Betts clusters by M. Jarrell in 2004.
c
c       We reserve all rights to this code.  Do not modify, distribute,
c       or run this code without our permission.
c
c       A very! large effort has been extended here, please do not add your 
c       name to this list unless you have dedicated a similar effort.
c
c       Parallelization details:
c       used mpi 1.2
c
c
c***************************************************************************
c To modify code to run on a 32 bit machine use the script 
c convert_64_to_32.sh
c***************************************************************************

        subroutine readin
c************************************************************
c       Readin the parameters and initialize some variables.
c************************************************************
        use Global
        implicit none
        include 'mpif.h'
#define SIMPLE_SPRNG
#define FLOAT_GEN
#define USE_MPI
#include "sprng_f.h"
c************************************************************ 
        character(8) :: waslin
        real(kind) :: usefrac
        integer i,ios
c   for sprng random number generator
        SPRNG_POINTER junkPtr
c   for mpi
        integer nprocin
c************************************************************ 


        if(myrank.eq.0) then
          abortf=0                         ! cannot stop only myrank=0
          open(unit=lub,file='bins.dat',status='unknown')
          open(unit=lus,file='sigma.dat',status='unknown')
#ifdef PUT_GAMAT
          open(unit=lug,file='gamat.dat',status='unknown',
     &         form='unformatted')
#endif
          open(unit=lui,file='indat',status='old')
          open(unit=lud,file='data',status='unknown')

          write(lud,*)'************************************************'
          write(lud,*)'PROGRAM DCA_HUBBARD'
          write(lud,*)'************************************************'

          read(lui,"(1x,a8)") waslin
          read(lui,*) nprocin,nl,cluster,nwn,ndim,iskip,umode,Nblock,
     &                iauto
     	  if (nl.gt.nwn) then
     	  	write (lud,*) 'nl cannot be larger than nwn'
     	  	abortf=1
     	  end if
          if (umode.lt.0.or.umode.gt.5) then
             write (lud,*) 'Unknown update mode umode=',umode
             abortf=1                         ! cannot stop only myrank=0
          end if
          if (umode.eq.0) then
             nblock=1 ! safety: set correct block size for single spin updates
          end if
          if (umode.eq.2.or.umode.eq.3) then
             ! set maximum block size
             if(nblock.gt.nl.or.nblock.lt.0) nblock=nl
          end if

          read(cluster,"(i3)",iostat=ios) Nc     ! should barf with characters
          ibetts=0
          if (ios.gt.0) then
            ibetts=1
            read(cluster,"(i2)",iostat=ios) Nc   ! ignore trailing charac.
            if (ios.gt.0) then
              read(cluster,"(i1)",iostat=ios) Nc ! ignore trailing charac.
              if (ios.gt.0) then
                write(lud,*) "Cannot read Nc from string ",cluster
                abortf=1                         ! cannot stop only myrank=0
              endif
            endif
          endif

          write(lud,*) 'Nc, ibetts= ',Nc,ibetts         ! check to be sure
          nt=nl*Nc
          if (ndim.eq.0.and.Nc.gt.1) then
            write(lud,*) 'infinite D for Nc > 1 not allowed!'
            stop
	  else if (ndim.eq.1 .and. Nc.eq.1) then
	    write(lud,*) 'ndim=1 for Nc=1 not allowed'
	    stop
          endif
          write(lud,*) '  procs      ndim'
          write(lud,"(3x,i4,6x,i4,3(6x,i2))") nprocin,ndim
          write(lud,*) '  iskip    umode   Nblock  iauto'
          write(lud,"(3x,i4,6x,i4,3(6x,i2))") iskip,umode,
     &                                        Nblock,iauto
          write(lud,*) '    nl     cluster      nwn       nt   '
          write(lud,"(3x,i4,6x,a4,6x,i4,6x,i4)") nl,cluster,nwn,nt
     

          read(lui,"(1x,a8)") waslin
          read(lui,*) warm,skip,meas,ntr,run,niter,usefrac,dfac,ifac,
     &               ntot
          write(lud,*) '   warm      skip      meas      ntr       run'
          write(lud,"(3x,i4,6x,i4,6x,i4,6x,i4,6x,i4)")
     &                                          warm,skip,meas,ntr,run
          write(lud,*) '  niter    usefrac     dfac     ifac      ntot'
          write(lud,"(3x,i4,5x,f6.4,4x,f6.4,4x,f6.4,5x,i4)") 
     &                             niter,usefrac,dfac,ifac,ntot
          

          read(lui,"(1x,a8)") waslin
          read(lui,*) iseed,isi,iso,isb,isd,ised,isg,iosft,npt,icond    
          write(lud,*) '  iseed      isi       iso       isb       isd'
          write(lud,"(1x,i9,3x,i4,6x,i4,6x,i4,6x,i4)")
     &                                          iseed,isi,iso,isb,isd
          write(lud,*) '   ised      isg     isosft      npt      icond'
          write(lud,"(3x,i4,6x,i4,6x,i4,6x,i4,6x,i4)") 
     &                                       ised,isg,iosft,npt,icond


          read(lui,"(1x,a8)") waslin
          read(lui,*)   ed,tprime,tperp,u,V0,beta,ftarget
          write(lud,*) '    ed      tprime    tperp       U    '
          write(lud,"(2x,f6.3,4x,f6.3,4x,f6.3,4x,f6.3)") 
     &                             ed,tprime,tperp,u
          write(lud,*) '    V0       beta    ftarget '
          write(lud,"(2x,f6.3,4x,f6.2,4x,f6.3)") V0,beta,ftarget

          write(lud,*)'************************************************'
          close(unit=lui)   
c
c         Set some parameters and write warnings if needed.
c        
          if(usefrac.gt.oner) 
     &      write(lud,*) 'usefrac too large => bounds errors'
          nuse=int(usefrac*nl)

          if(u.ge.zeror) then
            isu=+1
          else
            isu=-1
          end if
          if(abs(V0).ge.eps) then
            isv=1       ! include diagonal disorder of strength V0.
          else
            isv=0       ! The system is assumed to be ordered.
          end if
         
          nover=int(sqrt(float(4*ntot**ndim/Nc))/2) ! roughly the number of
                                            ! kt-points in coarse-graining cell
          write(lud,*) 'nover = ', nover
         
          write(lud,*)  '**********************************************'
          write(lud,*)  '* SWITCHES SET FOR:                          *'
          if(ntot.eq.0) then
           write(lud,*) '* Finite Sized Simulation                    *'
          else
            write(lud,*)'* DCA Simulation                             *'
          end if
          if(isv.eq.0) then
           write(lud,*) '* Ordered System                             *'
          else
            write(lud,*)'* Disordered System                          *'
          end if
          if(isu.eq.1) then
            write(lud,*)'* Positive U Model                           *'
          else
            write(lud,*)'* Negative U Model                           *'
          end if
          if(ibetts.eq.1) then
            write(lud,*)'* Betts Clusters Used                        *'
          else
            write(lud,*)'* Conventional (not Betts) clusters used     *'
          end if
          if(umode.eq.0) then
            write(lud,*)'* Using traditional single-spin updates      *'
          else if (umode.eq.1) then
            write(lud,*)'* Using delayed spin-updates                 *'
          else if (umode.eq.2.or.umode.eq.3) then
            write(lud,*)'* Using blocked spin-updates                 *'
          else if (umode.gt.3) then
            write(lud,*)'* Using CUDA                                 *'
          end if
          if(iskip.eq.0) then
            write(lud,*)'* Dont adjust skip                           *'
          else if (iskip.eq.1) then
            write(lud,*)'* Adjust skip, one autotime between meas     *'
          end if
          if(iauto.eq.0) then
            write(lud,*)'* Dont calculate autotime                    *'
          else if (abs(iauto).eq.1) then
            write(lud,*)'* Calculate spin autotime                    *'
          else if (abs(iauto).eq.2) then
            write(lud,*)'* Calculate pair autotime                    *'
          else if (abs(iauto).eq.3) then
            write(lud,*)'* Calculate charge autotime                  *'
          else if (abs(iauto).eq.4) then
            write(lud,*)'* Calculate G(K=0,iw0) autotime              *'
          else if (abs(iauto).eq.5) then
            write(lud,*)'* Calculate G(R=0,tau=0) autotime            *'
          end if
          if(iso.eq.0) then
            write(lud,*)'* Initialized with perturbation theory Sigma *'
          else if(iso.eq.1) then
            write(lud,*)'* Initialized with provided Sigma and ed     *'
          else if(iso.eq.2) then
            write(lud,*)'* Init. with provided Sigma, ed, and fields  *'
          else if(iso.eq.3) then
            write(lud,*)'* Initialized with provided Sigma            *'
          else if(iso.eq.4) then
            write(lud,*)'* Initialized with provided Sigma and fields *'
          end if
          if(isb.eq.0) then
            write(lud,*)'* No Binned Data Measured                    *'
          else
            write(lud,*)'* Binned Data Measured                       *'
          end if
          if(isg.eq.0) then
            write(lud,*)'* No 2-Particle Quantities Measured          *'
          else
            write(lud,*)'* 2-Particle Quantities Measured             *'
          end if
          if(iosft.eq.0) then
            write(lud,*)'* Oversampled FT uses perturbation theory    *'
          else
            write(lud,*)'* Oversampled FT uses the previous iteration *'
          end if
          if(icond.eq.0) then
            write(lud,*)'* Use PT for HF conditioning                 *'
          else
            write(lud,*)'* Use a/iwn +b/wn^2 for HF cond.             *'
          end if
          if(ised.eq.0) then
            write(lud,*)'* ed is fixed                                *'
          else
            write(lud,*)'* ed is adjusted to fix the filling          *'
          end if
          if(isd.eq.0) then
            write(lud,*)'* No damping is used in the calc. of Sigma   *'
          else
            write(lud,*)'* Damping is used in the calc. of Sigma      *'
          end if
          write(lud,*)  '**********************************************'
          
          
c         check to make sure that the number of processors is correct   !0
          if(nprocin.ne.nproc) abortf=1                                 !0
c         Ensure that brute force CG is used with Betts tables.         !0
          if(ibetts.eq.1.and.brufor.eq.0.and.abortf.eq.0) then          !0
            write(lud,*) 'abort in readin'
            abortf=1                                                    !0
            flush(lud)
          end if
        endif ! if(myrank.eq.0)

c now pass the variables to all processors
c       INTEGERS
        call MPI_BCAST(nl,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)        ! nl
        call MPI_BCAST(Nc,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)        ! Nc
        call MPI_BCAST(cluster,3,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr) ! cluster
        call MPI_BCAST(nwn,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! nwn
        call MPI_BCAST(ndim,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)      ! ndim
        call MPI_BCAST(umode,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)     ! umode
        call MPI_BCAST(Nblock,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)    ! Nblock
        call MPI_BCAST(iskip,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)     ! iskip
        call MPI_BCAST(iauto,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)     ! iauto
        call MPI_BCAST(nuse,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)      ! nuse
        call MPI_BCAST(warm,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)      ! warm
        call MPI_BCAST(skip,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)      ! skip
        call MPI_BCAST(meas,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)      ! meas
        call MPI_BCAST(ntr,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! ntr
        call MPI_BCAST(run,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! run
        call MPI_BCAST(niter,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)     ! niter
        call MPI_BCAST(ntot,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)      ! ntot
        call MPI_BCAST(nover,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)     ! nover
        call MPI_BCAST(ibetts,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)    ! ibetts
        call MPI_BCAST(iseed,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)     ! iseed
        call MPI_BCAST(isi,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! isi
        call MPI_BCAST(iso,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! iso
        call MPI_BCAST(isb,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! isb
        call MPI_BCAST(isd,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! isd
        call MPI_BCAST(ised,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)      ! ised
        call MPI_BCAST(isu,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! isu
        call MPI_BCAST(abortf,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)    ! abortf
        call MPI_BCAST(isg,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! isg
        call MPI_BCAST(isv,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! isv
        call MPI_BCAST(iosft,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)     ! iosft
        call MPI_BCAST(npt,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)       ! npt
        call MPI_BCAST(icond,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)     ! icond
        nt=nl*Nc
        if (ndim.eq.2) then
          N_sp=sqrt(float(Nc))
        else if (ndim.eq.3.and.ibetts.eq.0) then
          N_sp=sqrt(float(Nc/2))
        else if (ndim.eq.3.and.ibetts.eq.1) then
          N_sp=sqrt(float(Nc))
        end if
        NKmax=nl+(nl+1)/2 +1
        if (ndim.eq.0) then
           ngroup=8  ! since we will use the 2D tables
        else
        ngroup=1
        do i=1,ndim
          ngroup=ngroup*(ndim-i+1)  ! ndim factorial
        end do
        ngroup=ngroup*2**ndim       ! i.e. ngroup =2,8,48 for D=1,2,3
        endif

c       REALS
        call MPI_BCAST(usefrac,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)     ! usefrac
        call MPI_BCAST(dfac,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)        ! dfac
        call MPI_BCAST(ifac,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)        ! ifac
        call MPI_BCAST(ed,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)          ! ed
        call MPI_BCAST(u,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)           ! u
        call MPI_BCAST(beta,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)        ! beta
        call MPI_BCAST(ftarget,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)     ! ftarget
        call MPI_BCAST(tprime,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)      ! tprime
        call MPI_BCAST(V0,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)          ! V0
        call MPI_BCAST(tperp,1,MPI_REAL8,0,MPI_COMM_WORLD,ierr)       ! tperp
 

c       do we need to abort?
        write(abortstring,*) 'readin'
        call abortquerry
        
c       Now that we know the size of things, allocate the arrays
        call allocate_1p
        if(myrank.eq.0.and.iprint.ge.1) then
          write(lud,*) 'done allocate_1p'                               !0
        end if
c       The two-particle arrays are allocate in tables (after Ncw is
c       determined)

c       set up the random number library -- which gives an
c       independent random number sequence on each processor
c       http://www.ncsa.uiuc.edu/Apps/CMP/RNG/www/quick-start.html
c DO NOT EDIT: Makefile seds SPRNG 2.0->1.0 if necessary
c       junkPtr = init_sprng(iseed,SPRNG_DEFAULT)            !version 1.x SPRNG
        junkPtr = init_sprng(SPRNG_CMRG,iseed,SPRNG_DEFAULT) !version 2.0 SPRNG
c       call ranset(iseed+myrank)                            !CRAY
c       call srand(iseed+myrank)                             !HP or pgi

c       set some parameters
        temp=oner/beta
        iter=0
        if(myrank.eq.0.and.iprint.ge.1) then
		write(lud,*) 'done readin';flush(lud)
	endif
        return
        end
        
c                      123456789|123456789|123456789|123456789|123456789|


        subroutine sigmainit
c************************************************************
c       Initialize sigma in preparation for the first iteration
c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        integer :: i,j,n,m,i1
        real(kind) :: r1
        character(72) :: linemc
c************************************************************
        if(myrank.eq.0.and.iprint.ge.1) then
		write(lud,*) 'start sigmainit';flush(lud)
	endif
        isf=0
        s=0
        test=zeror        !convergence criteria

        if(iso.ge.1) then ! readin sigma from a file    
          if(myrank.eq.0) then                                        
c           do all readin from proc 0                                 !0
                                                                      !0
 2001       read(lus,1000) linemc                                     !0
            if(index(linemc(1:72),'Results').ne.0 ) then              !0
              read(lus,*) i1                                          !0
            else                                                      !0
              goto 2001                                               !0
            endif                                                     !0

 2002       read(lus,1000) linemc                                     !0
            if(index(linemc(1:72),'ed,u,beta').ne.0 ) then            !0
              read(lus,*) r1                                          !0
              if(iso.le.2) ed=r1                                      !0
            else                                                      !0
              goto 2002                                               !0
            endif                                                     !0
                                                                      !0
 2003       read(lus,1000) linemc                                     !0
            if(index(linemc(1:72),'nfocc').ne.0 ) then                !0
              if(index(linemc(1:72),'skip').ne.0.and.iskip.eq.1) then !0
                read(lus,*) nfocc,r1,skip                             !0
              else                                                    !0
                read(lus,*) nfocc                                     !0
              end if                                                  !0
            else                                                      !0
              goto 2003                                               !0
            endif                                                     !0
                                                                      !0
 2004       read(lus,1000) linemc                                     !0
            if(index(linemc(1:72),'sigma(').ne.0 ) then                !0
              do i=1,Ncw              !read data in the IW            !0
              do n=0,nwn                                              !0
                read(lus,*,end=997) m,m,sigma(n,i)                    !0
                if(n.lt.nwn) sigma(-n-1,i)=conjg(sigma(n,i))          !0
              end do                                                  !0
              end do                                                  !0
              do i=Ncw+1,Nc            ! fill in data outside IW      !0
              do n=0,nwn-1                                            !0
                sigma(n,i)=sigma(n,ickmap(i))                         !0
                sigma(-n-1,i)=conjg(sigma(n,i))                       !0
              end do                                                  !0
              end do                                                  !0
              write(lud,*) 'code init. with supplied sigma'           !0
            else if((iso.eq.2.or.iso.eq.4).and.                       !0
     &               index(linemc(1:72),'fields').ne.0.and.
     &               nl.eq.i1) then          !0
              read(lus,*) sgn                                         !0
              do i=1,nt                                               !0
                read(lus,*) j,s(i)                                    !0
              end do                                                  !0
              write(lud,*) 'code init. with supplied fields'          !0
              isf=1                                                   !0
              goto 2004                                               !0
            else                                                      !0
              goto 2004                                               !0
            endif                                                     !0
 997            continue                                              !0
          end if ! myrank.eq.0                                        !0
c         now send the data to everyone
          call MPI_BCAST(isf,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
          if(iskip.eq.1) 
     &      call MPI_BCAST(skip,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
          if(isf.eq.1) then
            call MPI_BCAST(s,nt,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
            call MPI_BCAST(sgn,1,
     &                     MPI_REAL8,
     &                     0,MPI_COMM_WORLD,ierr)
          end if  
          call MPI_BCAST(ed,1,
     &                   MPI_REAL8,
     &                   0,MPI_COMM_WORLD,ierr)
          call MPI_BCAST(nfocc,1,
     &                   MPI_REAL8,
     &                   0,MPI_COMM_WORLD,ierr)
          call MPI_BCAST(sigma,(2*nwn+1)*Nc,
     &                   MPI_COMPLEX16,
     &                   0,MPI_COMM_WORLD,ierr)
        else if(iso.eq.0) then           ! generate initial sigma from PT
          sigma=zeroc                    ! Initial guess for sigma
          nfocc=halfr*ftarget            ! and nf (needed for Hartree term)
          do i=1,npt
            call CoarseGrain(GcKf,sigma) ! make GcKf
            GcKfsc=GcKf/(oner+GcKf*sigma)! make GcKfsc used below
            call perturbation            ! use GcKfsc to calculate sigmapt
            sigma=sigmapt                ! to start over
          end do !i
          if(myrank.eq.0) write(lud,*) 'Code initialized with PT'      !0
        end if
 1000   format(a72)
 
        if(myrank.eq.0.and.iprint.ge.9) then
		write(lud,*) 'done sigmainit';flush(lud)
	endif
        return
        end subroutine sigmainit






        subroutine Header_bins
c*******************************************************************
c       Write out a header for the binned data
c*******************************************************************
        use Global
        implicit none
        include 'mpif.h'
c*****************************************************************
         integer :: n,rmeas
c*****************************************************************

        write(lub,*) 'Results from dca for nl,nwn,cluster,ndim='
        write(lub,*) nl,nwn,cluster,ndim
        write(lub,*) 'parameters follow'
        write(lub,*) 'warm,skip,meas,ntr,run,niter,nuse,dfac,ifac'
        rmeas=meas
        do n=1,niter-1
          rmeas=int(ifac*rmeas)
        end do
        write(lub,2) warm,skip,rmeas,ntr,nproc*run,niter,nuse,dfac,ifac
 2      format(i5,2x,i5,2x,i5,2x,i5,2x,i5,2x,i5,2x,i5,1x,f4.2,1x,f4.2)
        write(lub,*) 'random number, isb, isi, isd'
        write(lub,*) iseed,isb,isi,isd
        write(lub,*) '  ed    ,  tprime,    U   ,   beta  ,   V, tperp '
        write(lub,4) ed,tprime,u,beta,V0,tperp
 4      format(1x,f9.4,1x,f9.4,1x,f9.4,1x,f9.4,1x,f9.4,1x,f9.4)
        write(lub,*) ' '
        if(myrank.eq.0.and.iprint.ge.9) write(lud,*) 'done header'  !0
        return

        end subroutine Header_bins



        subroutine Header_Sigma

c************************************************************
c       Write Parameters to sigma.dat
c************************************************************
        use Global
        implicit none

c************************************************************ 
c************************************************************ 

        write(lus,*) 'Results for nl,nwn,filling,cluster,ntot,ndim='   
        write(lus,*) nl,nwn,2.0*nfocc,cluster,ntot,ndim                
        write(lus,*) 'ed,u,beta,iter,run,meas,tprime,V,tperp'      
        write(lus,*) ed,u,beta,niter,lrun,meas,tprime,V0,tperp     
        write(lus,*) 'nfocc=     <Sign>     skip'                           
        write(lus,*) nfocc,sgnf,skip                                    
                                       
        return
        end                           

        subroutine CoarseGrain(Gcg,sigmacg)   ! 1,2 and 3D VERSION
c**********************************************************************
c       Calculate the coarse-grained Green function, Gcg, corresponding
c       to the self energy sigmacg for the 1, 2 and 3D cluster.
c
c       To parallelize this block, we need a map from the process  
c       rank to the cluster site indices.  Each process gets its 
c       K-value, where its is determined by
c
c       its=Nc/nproc    nleft=Nc-nproc*its
c       if(myrank+1.le.nleft) its=its + 1
c       do i=1,its
c         ic = (myrank + 1) + nproc*(i-1)
c       end do
c
c**********************************************************************
        use Global
        use Dispersion
        implicit none
        include 'mpif.h'
#define SIMPLE_SPRNG
#define FLOAT_GEN
#define USE_MPI
#include "sprng_f.h"
c**************************************************************
        integer :: i,j,k,n,ic,its,kits,nleft,idim,iproc,ifail,
     &             nw,neval,maxcls
        real(kind) :: r1,r2,r3,r4,kx,ky,kz,Eps_k,tperp_k,Dkt,
     &                Deltak,alow(3),bhigh(3),finest(2),fac,
     &                absreq,relreq,absest(2),wrkstr(5000),
     &                ekin,wre,wim,zre,zim
        complex(kind) :: z,c1,c2,c3,zeta,
     &                   Gcg(-nwn:nwn,Nc),sigmacg(-nwn:nwn,Nc)
        logical :: flag
        external integ
c**************************************************************
        Gcg(:,:) = zeroc
        Dkt=halfr/real(nover,kind)
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
        its=Nc/nproc
        nleft=Nc-nproc*its
        if(myrank+1.le.nleft) its = its + 1
        
        if (ndim.eq.0.and.Nc.eq.1) then
c       infinite D: Do integration with Gaussian DOS 
c 
c                            +oo
c                    1        /                exp(-e^2)               sqrt(pi)
c       G(iwn) =  --------    | de  ------------------------------  =  -------- * w(z)
c                 sqrt(pi)   /       [iwn - ed - sigma(iwn)] - e          ii
c                            -oo
c
c      where z = iwn - ed - sigma(iwn)
c      by using the error-function.
c      This is implemented with the subroutine wofz.f.
c      Note: Nc has to be set to one!

          if (myrank.eq.0) then

            ekin=0.0_kind

            do n=0,nwn-1

              zeta = ii*wn(n) - ed - sigmacg(n,1)
              zim  = aimag(zeta)
              zre  = real(zeta)

              call wofz(zre,zim,wre,wim,flag)
              if (flag) then
                 write(6,*) "Problems with errfc at n = ", n
              stop 
              endif

              Gcg(n,1) = sqrt(pi)*cmplx(wim,-wre)

              ekin=ekin+twor*sqrt(pi)*(zre*wim+zim*wre)-twor

            enddo               ! n

             ekin=ekin*temp
             write(lud,*) "Kinetic energy: ", ekin 

          endif ! myrank=0


        else if(brufor.eq.0.and.ndim.ge.1) then  
c 
c       Do numeric integration using DCUHRE-adaptive quadrature package
c       The integration is done for the real and imaginary part individually.
c       The functions for these parts are evaluated in the functions integ(1)
c       and integ(2) respectively.
c           ____ 
c           \                           -w(n) + Im(sigma(iwn,K))                                
c Im =       \  -------------------------------------------------------------                      
c            / ( w(n) - Im(sigma(iwn,K)))^2 + (Eps_k+ed+Re(sigma(iwn,K)))^2                          
c           /___                                                                  
c             k
c
c
c           ____ 
c           \                      -Eps_k-ed-Re(sigma(iwn,K))                                   
c Re =       \  -------------------------------------------------------------                      
c            / ( w(n) - Im(sigma(iwn,K)))^2 + (Eps_k+ed+Re(sigma(iwn,K)))^2                          
c           /___                                                                  
c             k
c

           absreq=0.000001    !requested absolute error
           relreq=0.000001    !requested relative error
           maxcls=1000        !maximal allowed function evaluations
           nw=5000


          do kits=1,its
            ic = (myrank + 1) + nproc*(kits-1) ! 1 < ind <= Nc


c       Set the limits of integration according to dimensionality
          if(ndim.eq.1) then
            alow(1)=Kc(1,ic) - halfr*g(1,1)  !lower bound for x
            alow(2)=-pi                      !lower bound for y
            alow(3)=-pi                      !lower bound for z
            bhigh(1)=Kc(1,ic) + halfr*g(1,1) !upper bound for x
            bhigh(2)=pi                      !upper bound for y
            bhigh(3)=pi                      !upper bound for z
            fac = real(Nc,kind)/(twor*pi)**3 !Integral normalizatiopn factor

          else if(ndim.eq.2) then
            alow(1)= - halfr*sqrt(g(1,1)**2+g(1,2)**2) !lower bound for x
            alow(2)= - halfr*sqrt(g(2,1)**2+g(2,2)**2) !lower bound for y
            alow(3)=-pi                                !lower bound for z
            bhigh(1)= + halfr*sqrt(g(1,1)**2+g(1,2)**2)!upper bound for x
            bhigh(2)= + halfr*sqrt(g(2,1)**2+g(2,2)**2)!upper bound for y
            bhigh(3)=pi                                !upper bound for z
            fac=1/(twor*pi*sqrt(g(1,1)**2+g(1,2)**2) 
     &             *sqrt(g(2,1)**2+g(2,2)**2))   !Integral normalizatiopn factor

          else if(ndim.eq.3) then
            alow(1)=-halfr*sqrt(g(1,1)**2+g(1,2)**2+g(1,3)**2)!lower bound for x
            alow(2)=-halfr*sqrt(g(2,1)**2+g(2,2)**2+g(2,3)**2)!lower bound for y
            alow(3)=-halfr*sqrt(g(3,1)**2+g(3,2)**2+g(3,3)**2)!lower bound for z
            bhigh(1)=halfr*sqrt(g(1,1)**2+g(1,2)**2+g(1,3)**2)!upper bound for x
            bhigh(2)=halfr*sqrt(g(2,1)**2+g(2,2)**2+g(2,3)**2)!upper bound for y
            bhigh(3)=halfr*sqrt(g(3,1)**2+g(3,2)**2+g(3,3)**2)!upper bound for z
            fac =1/(sqrt(g(1,1)**2+g(1,2)**2+g(1,3)**2)
     &              *sqrt(g(2,1)**2+g(2,2)**2+g(2,3)**2)
     &              *sqrt(g(3,1)**2+g(3,2)**2+g(3,3)**2)) !Integral normalizatiopn factor
          end if   

cdir$ preferstream
            do n=0,nwn-1
              curtdata(2)=aimag(sigmacg(n,ic))  !set global as needed for function evaluation
              curtdata(1)=real(sigmacg(n,ic))
              curtdata(3)=wn(n)
              curtdata(4)=ic             
              call DCUHRE(3, 2, alow, bhigh,0, maxcls, integ,
     &           absreq, relreq, 0, nw, 0, finest, absest, neval,
     &           ifail, wrkstr)  !perform integration
            if(ifail.eq.1) write(6,*) 'Desired accuracy not obtained', 
     &            'Continuing with error=', absest 
            if(ifail.gt.1) write(6,*)'Numeric integration failed error',
     &            ifail
              Gcg(n,ic)=fac*(finest(2)+ii*finest(1))
            end do
          end do

          else if(ndim.eq.1.and.brufor.eq.1) then
c 
c         ONE DIMENSIONAL CASE - "brute force"
c         The integral over kt is performed numerically as is the
c         sum over the environment 
c
          do kits=1,its
            ic = (myrank + 1) + nproc*(kits-1)       ! 1 < ind <= Nc
            do i=-nover+1,nover
            do j=-nover+1,nover
            do k=-nover+1,nover
              kx = Kc(1,ic) + Dkt*g(1,1)*(real(i,kind)-halfr)
              ky = Dkt*twor*pi*(real(j,kind)-halfr)
              kz = Dkt*twor*pi*(real(k,kind)-halfr)
              Eps_k   = -halfr*(cos(kx) + tprime*cos(twor*kx))
     &                  -halfr*tperp*(cos(ky)+cos(kz))
cdir$ preferstream
              do n=0,nwn-1
                z = ii*wn(n) - Eps_k - ed - sigmacg(n,ic)
                Gcg(n,ic)=Gcg(n,ic)+ oner/z
              end do
            end do !k
            end do !j
            end do !i    
            Gcg(:,ic)=Gcg(:,ic)/real((2*nover)**3,kind)
          end do
        else if(ndim.eq.2.and.brufor.eq.1) then
c 
c         TWO DIMENSIONAL CASE
c

          do kits=1,its
            ic = (myrank + 1) + nproc*(kits-1)       ! 1 < ind <= Nc
            do i=1,nwsc
              kx=Kc(1,ic) + kt(1,i)
              ky=Kc(2,ic) + kt(2,i)
              Eps_k   = -halfr*(cos(kx)+cos(ky)) -
     &                   tprime*(cos(kx)*cos(ky)-oner)
c              tperp_k = -halfr*tperp*(cos(kx)-cos(ky))**2  !bilayer
              tperp_k = -halfr*tperp
cdir$ preferstream
	      if (Nc.eq.1 .and. ntot.eq.0) Eps_k = 0
              do n=0,nwn-1
                z = ii*wn(n) - Eps_k - ed - sigmacg(n,ic)
                c1=tperp_k + z
                c2=sqrt((z-tperp_k)/c1)
                Gcg(n,ic)=Gcg(n,ic)+ oner/(c1*c2)
              end do
            end do !i
            Gcg(:,ic)=Gcg(:,ic)/real(nwsc,kind)
          end do

        else if(ndim.eq.3.and.brufor.eq.1) then
c
c         THREE DIMENSIONAL CASES (ndim=3)
c         The sums over kx,ky and kz are performed numerically.
c
          do kits=1,its
            ic = (myrank + 1) + nproc*(kits-1)       ! 1 < ind <= Nc
            do i=1,nwsc
              kx=Kc(1,ic) + kt(1,i)
              ky=Kc(2,ic) + kt(2,i)
              kz=Kc(3,ic) + kt(3,i)
              Eps_k=Epsk_nD((/kx,ky,kz/))
cdir$ preferstream
              do n=0,nwn-1
                z = ii*wn(n) - Eps_k - ed - sigmacg(n,ic)
                Gcg(n,ic)=Gcg(n,ic)+ oner/z
              end do
            end do
            Gcg(:,ic)=Gcg(:,ic)/real(nwsc,kind)
          end do 
        end if  ! ndim = 1,2, 3

        do kits=1,its
          ic = (myrank + 1) + nproc*(kits-1)       ! 1 < ind <= Nc
          do n=0,nwn-1
            Gcg(-n-1,ic)=conjg(Gcg(n,ic))
            if(aimag(Gcg(n,ic)).gt.zeror) then
              write(6,*) 'CAUSALITY VIOLATION, n,ci=',n,ic  !trap for causality
            end if
          end do
        end do

c       Now on every processor, pass the information around
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
        do ic=1,Nc
          iproc = ic - 1 - nproc*((ic-1)/nproc)
          call MPI_BCAST(Gcg(-nwn,ic),2*nwn+1,
     &                   MPI_COMPLEX16,
     &                   iproc,MPI_COMM_WORLD,ierr)
        end do


        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
        
        return
        end subroutine CoarseGrain

        subroutine FT_Kf_to_Kt(GcgKf,GcgKt,itype)
c**********************************************************************
c       Fourier Transform GcgKf(iwn,ic) to GcgKt(l,ic).
c       Conditioning will be used to get the high-frequencies
c       right. If itype=0  then No Hartree term is used, as
c       GcgKf is assumed to be cluster excluded
c       
c       
c**********************************************************************
        use Global
        implicit none
c**************************************************************
        integer :: n,ic,itype,j
        real(kind) :: r1,nd,ta,goo,GcgKt(-nl:nl,Nc)
        complex(kind) :: GcgKf(-nwn:nwn,Nc),csum1
c**************************************************************

c       First, the time part, the result is real due to inversion symm.
c       Conditioning is used
c
c       g0(t,x)= [1-nd] exp(-t*x)  t>0   nd=1/(exp(beta*x)+1)
c       g0(t,x)= [-nd ] exp(-t*x)  t<0   (note the wrong sign)
c
c       g0(iwn,x) = 1/(iwn-x)
c
c       G(t) = T SUM G(iwn) exp(-iwn t)  AGD pp.120-125
c
        do ic=1,Nc
          if(itype.eq.0) then
            r1=ed + Epsbar(ic) 
          else if(itype.eq.1) then
            r1=ed + Epsbar(ic) + u*(nfocc-halfr)
          else
            write(lud,*) 'wrong itype in FT_Kf_to_Kt'
          end if
c          
          do j=-nl,nl
            ta=dt*j
            if     (j.ge.0.and.r1.ge.zeror)then  
              goo=exp(-ta*r1)/(oner+exp(-beta*r1))
            else if(j.ge.0.and.r1.lt.zeror)then  
              goo=exp((beta-ta)*r1)/(oner+exp(beta*r1))
            else if(j.lt.0.and.r1.ge.zeror)then
              goo=-exp(-(beta+ta)*r1)/(oner+exp(-beta*r1))
            else if(j.lt.0.and.r1.lt.zeror)then
              goo=-exp(-ta*r1)/(oner+exp(beta*r1))
            else if(myrank.eq.0) then
              write(lud,*) 'WOOPS'
            end if
            csum1=zeroc
            do n=-nwn,nwn-1
              csum1=csum1+exp(-ii*wn(n)*ta)*
     &              (oner/(ii*wn(n)-r1)-GcgKf(n,ic))
            end do
            GcgKt(j,ic) = csum1*temp + goo
          end do
        end do
        
        return
        end subroutine FT_Kf_to_Kt

        subroutine Calculate_ed(chaclu,sig_nfocc)
c************************************************************
c       Recalculate ed to keep the filling fixed
c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************ 
        real(kind) :: chaclu,sig_nfocc,edold,r1
c************************************************************ 

        if(myrank.eq.0) then
          write(lud,"(' ed=',f9.6,' fill=',f9.6)") ed,twor*nfocc        !0
c                                                                       !0
c         RECALCULATE ed                                                !0
c                                                                       !0
          if(ised.eq.1.and.iter.ge.2                                    !0
     &       .or.                                                       !0
     &       ised.eq.2.and.iter.ge.1                                    !0
     &       .or.                                                       !0
     &       ised.eq.3) then                                            !0
c           use the charge susceptibility to calculate a new ed         !0
            chaclu=max(chaclu,real(0.3,kind))   ! set lower bounds      !0
            sig_nfocc=max(sig_nfocc,0.0001_kind)! to prevent NaN's      !0
            r1=exp(-halfr*((twor*nfocc-ftarget)/(twor*sig_nfocc))**2)   !0
            edold=ed                                                    !0
            ed=r1*ed+ (oner-r1)*(ed + (twor*nfocc-ftarget)/chaclu)      !0
            ed=dfac*ed+(oner-dfac)*edold                                !0
            write(lud,405) ed,twor*nfocc,twor*sig_nfocc                 !0
 405        format('ed=',f9.6,' filling=',f9.6,' +/- ',f9.6)            !0
          end if                                                        !0
                                                                        !0
        end if
        call MPI_BCAST(ed,1,
     &                 MPI_REAL8,
     &                 0,MPI_COMM_WORLD,ierr)


        call MPI_BCAST(nfocc,1,
     &                 MPI_REAL8,
     &                 0,MPI_COMM_WORLD,ierr)

        return
        end



        subroutine integ(ndims, x, nfun, f)
c       This subroutine provides the function for the integration of G
c       in the CoarsGrain subroutine
        use Global
        implicit none
        integer ::  ndims, nfun,ik
        real(kind) :: x(3),xp(3),f(2),w_n,resig,imsig,epl,
     &             ug(ndim)
        w_n=curtdata(3)
        resig=curtdata(1)
        imsig=curtdata(2)
        ik=curtdata(4)   

        if(ndim.eq.1) then 
          epl=- halfr*(cos(x(1))+tprime*cos(twor*x(1)))
     &          - halfr*tperp*(cos(x(2))+cos(x(3)))


        else if(ndim.eq.2) then 
          ug(1)=sqrt(g(1,1)**2+g(1,2)**2) !Length of g vector
          ug(2)=sqrt(g(2,1)**2+g(2,2)**2)
          xp(1)=(x(1)+Kc(1,ik))*g(1,1)/ug(1)
     &           +(x(2)+Kc(2,ik))*g(2,1)/ug(2) 
          xp(2)=(x(1)+Kc(1,ik))*g(1,2)/ug(1)
     &           +(x(2)+Kc(2,ik))*g(2,2)/ug(2)
          epl = -halfr*(cos(xp(1))+cos(xp(2))) 
     &                  - tprime*(cos(xp(1))*cos(xp(2))-oner)
     &                  -halfr*tperp


        else if(ndim.eq.3.and.ibetts.eq.0) then
          xp(1)=(x(1)+Kc(1,ik))*g(1,1)+(x(2)+Kc(2,ik))*g(2,1)
     &             +(x(3)+Kc(3,ik))*g(3,1)
          xp(2)=(x(1)+Kc(1,ik))*g(1,2)+(x(2)+Kc(2,ik))*g(2,2)
     &             +(x(3)+Kc(3,ik))*g(2,3)
          xp(3)=(x(1)+Kc(1,ik))*g(1,3)+(x(2)+Kc(2,ik))*g(2,3)
     &             +(x(3)+Kc(3,ik))*g(3,3)

          epl = -halfr*(cos(xp(1))+cos(xp(2))+cos(xp(3)))
     &      - tprime*(cos(xp(1))*cos(xp(2))
     &      + cos(xp(1))*cos(xp(3))+  cos(xp(2))*cos(xp(3)))
        else if(ndim.eq.3.and.ibetts.eq.1) then
          write(0,*) 'cannot do CG by integration if ibetts is 1'
          stop
        end if

        f(1) =  (-w_n+imsig)/((w_n-imsig)**2+(ed + epl+resig)**2)
        f(2) = (-ed-epl-resig)/((w_n-imsig)**2+(ed+epl+resig)**2)

        end
